"""
Created on Wed Mar 30 12:13:39 2022

@author: Haotian Teng
"""
import torch
import itertools
from torch.nn.functional import normalize,one_hot
import numpy as np
from typing import List,Union
class GaussianEmissions(torch.nn.Module):
    def __init__(self, means:np.array,cov:np.array, trainable:np.array = None, fix_cov = True):
        """
        The Gaussian emission module to generate .

        Parameters
        ----------
        means : np.array, shape [N,K]
             The means of the N states, K is the number of features.
        cov : np.array, shape [N,K]
            The diagonal covariance matrix of the N states and K feature.
        trainable : np.array, [N]
            To mark if the state is trainable. The default is None, will train
            all the states.
        fix_cov : bool
            If we want to fix the covariance during training.

        """
        super().__init__()
        self.N, self.K = means.shape
        self.means = torch.nn.Parameter(torch.from_numpy(means).float())
        self.trainable = torch.from_numpy(trainable).bool() if trainable is not None else torch.Tensor([1]*self.N).bool()
        self.cov = torch.nn.Parameter(torch.from_numpy(cov).float(),requires_grad = not fix_cov)
        self.certainty = torch.nn.Parameter(torch.zeros((self.N,self.K)),requires_grad = False)
        self.momentum_mean = torch.nn.Parameter(torch.zeros((self.N,self.K)),requires_grad = False)
        self.momentum_mean_square = torch.nn.Parameter(torch.zeros((self.N,self.K)),requires_grad = False)
        self.global_step = torch.nn.Parameter(torch.tensor([1]*self.N),requires_grad = False)
        self.cache = None #The cache when Baum-Welch is used.
        self.epsilon = 1e-6 # A small number to prevent numerical instability
        self.fix_cov = fix_cov
        if self.trainable is not None:
            self.means.register_hook(lambda grad: grad*self.trainable[:,None].to(self.device))
    
    def reinitialize_means(self, means:torch.tensor):
        """
        Reinitialize the means, this function is used to update the means of
        the modified kmers.

        Parameters
        ----------
        means : torch.tensor
            The new means.
        """
        with torch.no_grad():
            self.means.copy_(means)
            if hasattr(self,"global_step"):
                self.global_step = 1
    
    def forward(self, x:torch.Tensor) -> torch.Tensor:
        """
        Get the emission probability given a observation at time t.

        Parameters
        ----------
        x : torch.Tensor with shape [*,K]
            The observation value at certain time point, B is the batch size
            and K is the number of feature.

        Returns
        -------
        posterior: torch.Tensor with shape [*,N], return the log posterior 
        probability P(z|x), N is the number of states.
        """
        return -0.5*((1/self.cov*(self.means-x.unsqueeze(dim = -2))**2).sum(dim = -1)+torch.log(self.cov).sum(dim = 1)+self.K*np.log(2*np.pi))
    
    @torch.no_grad()
    def clamp_cov(self):
        """
        Clamp the covariance to positive number

        """
        self.cov[:] = torch.clamp(self.cov,min = self.epsilon)
    
    @torch.no_grad()        
    def update_cov(self,
                   method = "max",
                   exploration = 1.):
        """
        Set the uniform covariance according to the uncertainty, this function
        is useful when the covariance need to be reset during training.

        """
        if method == "max":
            cov = torch.max(self.certainty[self.trainable])
        else:
            raise NotImplementedError("Other covariance reestimation methods \
                                      beside max is not implemented .")
        print("Rescale the covariance to %.2f"%(cov))
        self.cov[:] = cov*exploration
    
    def _initialize_sufficient_statistics(self):
        self.cache = {"means":torch.zeros((self.N,self.K),device = self.device),
                      "covs":torch.zeros((self.N,self.K),device = self.device),
                      "posterior":torch.zeros(self.N,device = self.device)}
        
    @torch.no_grad()
    def _accumulate_sufficient_statistics(self, x:torch.Tensor, gamma:torch.Tensor):
        """
        Accumulate the sufficient statistics for the emission model given batch
        of the observation and posterior probability of the hidden variable.

        Parameters
        ----------
        x : torch.Tensor with shape [B,T,K]
            A batch of the observation time-series signal.
        gamma : torch.Tensor with shape [B,T,N]
            The posterior probability of the given observation batch.
        """
        if self.cache is None:
            self._initialize_sufficient_statistics()
        self.cache["means"] += (x.unsqueeze(dim = 2)*gamma.unsqueeze(dim = 3)).sum(dim = (0,1))
        self.cache["covs"] += (((x.unsqueeze(dim = 2)-self.means)**2)*gamma.unsqueeze(dim = 3)).sum(dim = (0,1))
        self.cache["posterior"] += gamma.sum(dim = (0,1))
    
    @torch.no_grad()
    def update_parameters(self, 
                          lr:float = 0.1, 
                          weight_decay:float = 0.0,
                          momentum:float = 0.9,
                          second_momentum:float = 0.999,
                          eps = 1e-8):
        """
        Update the parameter from the accumulated sufficient statistics, this 
        method was used if the model is updated using the Baum-Welch algorithm
        instead of direct gradient descent.
        
        Parameters
        ----------
        moving_average: float, optional
            The factor a of the moving average, the parameter will be updated 
            by aP_{t-1} + (1-a)P_t, default is 0 which means no smoothing will
            be used.

        """
        g_mean = self.means - self.cache["means"]/(self.cache["posterior"]+self.epsilon).unsqueeze(dim = 1)
        self.trainable = self.trainable.to(self.device)
        update_mask = torch.logical_and(self.trainable,self.cache["posterior"]>0)
        not_update_mask = torch.logical_not(update_mask)
        if weight_decay > 0:
            g_mean += weight_decay * self.means
        if momentum != 0:
            if not hasattr(self,"momentum_mean"):
                self.momentum_mean = torch.nn.Parameter(torch.zeros((self.N,self.K),device = self.device),requires_grad = False)
                self.momentum_mean_square = torch.nn.Parameter(torch.zeros((self.N,self.K),device = self.device),requires_grad = False)
                self.global_step = torch.nn.Parameter(torch.tensor([1]*self.N,device = self.device),requires_grad = False)
            new_momentum = momentum*self.momentum_mean + (1-momentum)*g_mean
            new_second_momentum = second_momentum*self.momentum_mean_square + (1-second_momentum)*(g_mean**2)
            # new_momentum[not_update_mask] = self.momentum_mean[not_update_mask]
            # new_second_momentum[not_update_mask] = self.momentum_mean_square[not_update_mask]
            self.momentum_mean[update_mask] = new_momentum[update_mask]
            self.momentum_mean_square[update_mask] = new_second_momentum[update_mask]
            mmt_mean_hat = self.momentum_mean/(1-momentum**self.global_step[:,None])
            mmt_mean_sq_hat = self.momentum_mean_square/(1-second_momentum**self.global_step[:,None])
            g_mean = mmt_mean_hat/(torch.sqrt(mmt_mean_sq_hat)+eps)
            self.global_step[update_mask] += 1
        g_mean[not_update_mask] = 0
        self.means -= lr*g_mean
        new_certainty = momentum*self.certainty + (1-momentum)*self.cache["covs"]/(self.cache["posterior"]+self.epsilon).unsqueeze(dim = 1)/(1-momentum**self.global_step[:,None])
        self.certainty[update_mask] = new_certainty[update_mask]
        if not self.fix_cov:
            self.cov = self.certainty
        self._initialize_sufficient_statistics()
        return g_mean

class RHMM(torch.nn.Module):
    def __init__(self, 
                 emission_module:torch.nn.Module, 
                 transition_module:torch.nn.Module = None,
                 device:str = None,
                 normalize_transition:bool = False,
                 transition_operation:str = "sparse",
                 index_mapping:List[np.array] = None):
        """
        The Restricted hidden Markov model (RHMM), like the Non-homogenous HMM,
        that is the transition matrix of the model is varying among time, but 
        different from NHMM is that the transition matrix is restricted to 
        small amount of transitions for specific time t.

        Parameters
        ----------
        emission_module: torch.nn.Module
            The emission model to use, need to be a Module itself.
        transition_module: torch.nn.Module, optional
            The transition model, can be None as the transition matrix can be
            provided externally.
        device : str
            The device to run the model, can be cpu or cuda:0
        normalize_transition : bool, optional
            If we want to normalize the transition matrix, default is False.
        transition_operation: str, optional default is sparse
            The way of transition multiplication, default is sparse using
            sparse matrix operation, can be dense or compact.
        index_mapping: List[np.array], optional
            A list of index mapping matrix [from_map_matrix,to_map_matrix] can 
            be given if the transition matrix is given in a compact form.
        """
        super(RHMM, self).__init__()
        self.add_module("emission",emission_module)
        self.n_states = self.emission.N
        self.device = self._get_device(device)
        self.emission.device = self.device
        self.emission.to(device)
        self.epsilon = 1e-6
        self.log_epsilon = torch.log(torch.tensor(self.epsilon)) 
        self.normalization = normalize_transition
        possible_operations = ["sparse","dense","compact"]
        max_funcs = {"sparse":self.transition_max_sparse,
                               "dense":self.transition_max,
                               "compact":self.transition_max_compact}
        forward_funcs = {"sparse":self.transition_prob_sparse,
                                   "dense":self.transition_prob,
                                   "compact":self.transition_prob_compact}
        backward_funcs = {"sparse":self.transition_backward_sparse,
                          "dense":self.transition_backward,
                          "compact":self.transition_backward_compact}
        if transition_operation not in possible_operations:
            raise ValueError("Transition operation type can only be "+ ','.join(possible_operations)+" but %s is given"%(transition_operation))
        self.transition_operation = transition_operation
        if self.transition_operation == "compact":
            if index_mapping is None:
                raise ValueError("Index mapping is required for compact transition operation.")
            elif len(index_mapping) != 2:
                raise ValueError("Require two index mapping matrix, map_from, map_to.")
            else:
                source_mapping, target_mapping = index_mapping
                self.source_mapping = torch.from_numpy(source_mapping).long().to(device)
                self.target_mapping = torch.from_numpy(target_mapping).long().to(device)
        if transition_module:
            self.add_module("transition",transition_module)
        self.to(self.device)
        
    def forward(self, observation:torch.Tensor, 
                duration:torch.Tensor, 
                transition:List,
                start_prob:torch.Tensor = None) -> np.array:
        """
        The forward algorithm for gradient descent training.
        
        Parameters
        ----------
        observation : torch.Tensor with shape [B,T,K]
            Tensor gives a batch of observatipns where B is the batch size, 
            T is the time and K is the feature.
        duration: torch.Tensor with shape [B]
            Gives the duration of each observation.
        transition : A length L list and each element is torch.sparse_coo_tensor 
            with shape [B,N,N] gives the transition matrix at certain batch id, 
            where N is the number of states, first N index refers to the source 
            and second index refers to the target.
        start_prob: torch.Tensor with shape [B,N], optional.
            Give the start probability for the signal in batches, if None a 
            uniform distribution will be used.
        """
        log_alpha = self._forward(observation, transition,start_prob)
        log_sums = log_alpha.logsumexp(dim=2)
        log_probs = torch.gather(log_sums, 1, duration.view(-1,1) - 1)
        return log_probs
    
    def _forward(self, observation:torch.Tensor, 
                transition:np.array,
                start_prob:torch.Tensor) -> np.array:
        """
        Caculate the forward messages given a observation and transtion matrix,
        the log probability is returned.

        Parameters
        ----------
        observation : torch.Tensor with shape [B,T,K]
            Tensor gives a batch of observatipns where B is the batch size, 
            T is the time and K is the feature.
        transition : A length L list and each element is torch.sparse_coo_tensor 
            with shape [B,N,N] gives the transition matrix at certain batch id, 
            where N is the number of states, first N index refers to the source 
            and second index refers to the target.
        start_prob: torch.Tensor with shape [B,N]
            Give the start probability for the signal in batches.
            
        Returns
        -------
        log_alpha: torch.Tensor with shape [B,T,N]
            The log forward probability

        """
        batch_size,T,K = observation.shape     
        log_alpha = torch.zeros((batch_size,T,self.n_states),device = self.device)
        if start_prob is None:
            if transition[0].layout == torch.strided:
                start_prob = torch.sum(transition[0],dim = 2) + self.epsilon
            elif transition[0].layout == torch.sparse_coo:
                start_prob = torch.sparse.sum(transition[0],dim = 2).to_dense() + self.epsilon
            else:
                raise TypeError("Transition tensor need to be either dense or coo sparse tensor.")
        log_alpha[:,0,:] = self.emission(observation[:,0,:]) + start_prob.log()
        emission = self.emission(observation)
        for t in np.arange(1,T):
            if self.transition_operation == "sparse":
                assert not self.normalization, "Can't enable sparse operation will normalization is on."
                #TODO: implement nomalization for sparse operation.
                log_alpha[:,t,:] = emission[:,t,:] + self.transition_prob_sparse(transition[t],log_alpha[:,t-1,:])
            elif self.transition_operation == "dense":
                curr_transition = self.to_dense(transition[t])
                log_alpha[:,t,:] = emission[:,t,:] + self.transition_prob(curr_transition,log_alpha[:,t-1,:],normalization=self.normalization)
                del curr_transition
            elif self.transition_operation == "compact":
                n_batch,n_states,b2 = transition[t].shape
                curr_transition = transition[t][:,:,int(b2/2):]
                log_alpha[:,t,:] = emission[:,t,:] + self.transition_prob_compact(curr_transition,log_alpha[:,t-1,:],normalization=self.normalization)
                del curr_transition
            torch.cuda.empty_cache()
        return log_alpha
    
    def _backward(self, 
                  observation:torch.Tensor, 
                  transition:np.array,
                  duration:torch.Tensor,) -> np.array:
        """
        Calculate the backward messages given a observation and transition matrix
        Same parameters as the _forward method.
        
        Parameters
        ----------
        observation : torch.Tensor with shape [B,T,K]
            Tensor gives a batch of observatipns where B is the batch size, 
            T is the time and K is the feature.
        transition : A length L list and each element is torch.sparse_coo_tensor 
            with shape [B,N,N] gives the transition matrix at certain batch id, 
            where N is the number of states, first N index refers to the source 
            and second index refers to the target.
        duration: torch.Tensor with shape [B]
            Gives the duration of each observation.
        
        Returns
        -------
        log_beta: torch.Tensor with shape [B,T,N]
            The log backward probability.
        """
        batch_size,T,K = observation.shape
        log_beta = torch.zeros((batch_size,T,self.n_states),device = self.device)
        log_beta[:,duration-1,:] = 1./self.n_states
        for t in np.arange(T-2,-1,-1):
            batch_mask = t<(duration-1)
            if self.transition_operation == "sparse":
                if self.normalization:
                    raise ValueError("Can't enable sparse operation will normalization is on.")
                update = self.transition_backward_sparse(transition[t],
                                                         self.emission(observation[:,t+1,:]),
                                                         log_beta[:,t+1,:])
                log_beta[batch_mask,t,:] = update[batch_mask,:]
            elif self.transition_operation == "dense":
                curr_transition = self.to_dense(transition[t])[batch_mask,:,:]
                log_beta[batch_mask,t,:] = self.transition_backward(curr_transition,
                                                                    self.emission(observation[batch_mask,t+1,:]),
                                                                    log_beta[batch_mask,t+1,:],
                                                                    normalization = self.normalization)
                del curr_transition
            elif self.transition_operation == "compact":
                n_batch,n_states,b2 = transition[t].shape
                curr_transition = transition[t][batch_mask,:,:int(b2/2)]
                log_beta[batch_mask,t,:] = self.transition_backward_compact(curr_transition,
                                                                            self.emission(observation[batch_mask,t+1,:]),
                                                                            log_beta[batch_mask,t+1,:],
                                                                            normalization = self.normalization)
                del curr_transition
            torch.cuda.empty_cache()
        return log_beta
    
    def expectation(self, observation:torch.Tensor, 
                    duration:torch.Tensor,
                    transition:np.array,
                    start_prob:torch.Tensor = None) -> torch.Tensor:
        """
        Calculate the posterior probability P(z_i|X) using forward-backward 
        algorithm.
        
        Returns
        -------
        gamma: torch.Tensor with shape [B,T,N]
            The normalized posterior probability P(z_i|X)
        """
        batch_size,T,K = observation.shape
        log_alpha = self._forward(observation,transition,start_prob)
        log_beta = self._backward(observation,transition,duration)
        log_gamma = log_alpha + log_beta
        # print("log_alpha:",log_alpha)
        # print("log_beta:",log_beta)
        gamma = torch.softmax(log_gamma,dim = 2)
        if torch.sum(gamma.sum(dim = 2)==0):
            raise ValueError("gamma has pure 0 entry.")
        mask = torch.arange(start = 1, end = T+1,device = self.device).repeat(batch_size,1)>(duration.unsqueeze(dim = 1))
        gamma[mask] = 0
        # print("gamma: ",gamma)
        return gamma
    
    def sampling(self,
                 path:torch.Tensor,
                 certainty:float = None):
        """
        Sampling a observation given the path

        Parameters
        ----------
        path : torch.Tensor, shape [B,T]
            A batch of the kmer sequence to sampling from.
        certainty: float, optional, default is None.
            A uniform certainty to use to generate the sample, if None, the
            uncertainty parameter of the emission model will be used.
        Returns
        -------
        A sampled signal with shape [B,T,K].

        """
        B,T = path.shape
        std = self.emission.uncertainty[path] if certainty is None else certainty
        return self.emission.means[path]+std*torch.rand((B,T,self.emssion.K))

    def viterbi_decode(self, observation:torch.Tensor, 
                       duration:torch.Tensor, 
                       transition:List,
                       start_prob:torch.Tensor = None) -> np.array:
        """
        Calculate the viterbi path.
        
        Parameters
        ----------
        observation : torch.Tensor with shape [B,T,K]
            Tensor gives a batch of observatipns where B is the batch size, 
            T is the time and K is the feature.
        duration: torch.Tensor with shape [B]
            Gives the duration of each observation.
        transition : A length L list and each element is torch.sparse_coo_tensor 
            with shape [B,N,N] gives the transition matrix at certain batch id, 
            where N is the number of states, first N index refers to the source 
            and second index refers to the target.
        start_prob: torch.Tensor with shape [B,N], optional.
            Give the start probability for the signal in batches, if None a 
            uniform distribution will be used.
        """
        batch_size,T,K = observation.shape     
        path = torch.zeros((batch_size,T,self.n_states),dtype = torch.long, device = self.device)
        log_delta = torch.zeros((batch_size,T,self.n_states),device = self.device)
        viterbi_path = torch.zeros((batch_size,T),dtype = torch.long,device = self.device)
        if start_prob is None:
            if transition[0].layout == torch.strided:
                start_prob = torch.sum(transition[0],dim = 2) + self.epsilon
            else:
                start_prob = torch.sparse.sum(transition[0],dim = 2).to_dense() + self.epsilon
        log_delta[:,0,:] = self.emission(observation[:,0,:]) + start_prob.log()
        for t in np.arange(1,T):
            if self.transition_operation == "sparse":
                max_prob,max_idx = self.transition_max_sparse(transition[t], log_delta[:,t-1,:])
            elif self.transition_operation == "dense":
                curr_transition = self.to_dense(transition[t])
                max_prob, max_idx = self.transition_max(curr_transition, log_delta[:,t-1,:],normalization=self.normalization)
                del curr_transition
            elif self.transition_operation == "compact":
                n_batch,n_states,b2 = transition[t].shape
                curr_transition = transition[t][:,:,int(b2/2):]
                max_prob, max_idx = self.transition_max_compact(curr_transition, log_delta[:,t-1,:],normalization=self.normalization)
                del curr_transition
            log_delta[:,t,:] = self.emission(observation[:,t,:]) + max_prob
            path[:,t,:] = max_idx
            torch.cuda.empty_cache()
        viterbi_path.scatter_(1,(duration-1).unsqueeze(1),torch.argmax(log_delta[np.arange(batch_size),duration-1,:],dim = 1,keepdim = True))
        for t in np.arange(T-2,-1,-1):
            mask = duration > (t+1)
            viterbi_path[mask,t] = torch.gather(path[:,t+1,:],dim = 1,index = viterbi_path[:,t+1][:,None])[mask[:,None]]
        return viterbi_path,log_delta
    
    def transition_backward(self,
                            transition:torch.Tensor,
                            log_emission:torch.Tensor,
                            log_beta:torch.Tensor,
                            normalization:bool = True):
        """
        Calculate the step backward log probability.
        log beta_t^k = log sum_i exp{log T_t^{k,i} + log e^i(x_{t+1})+log beta_{t+1}^i}

        Parameters
        ----------
        transition : torch.Tensor shape [B,N,N]
            The transition matrix at current time point.
        log_emission : torch.Tensor shape [B,N]
            The log emission probability at time point t+1.
        log_beta : torch.Tensor shape [B,N]
            The t+1 log beta posterior probability.
        normalization : bool, opotional
            If we want to normalize the transition matrix.

        Returns
        -------
        The log beta posterior probabtiliy at current time point t.

        """
        if normalization:
            log_transition = normalize(transition,p=1,dim=2).log()
        else:
            log_transition = transition.log()
        elementwise_sum = log_transition + log_emission.unsqueeze(dim = 1) + log_beta.unsqueeze(dim = 1)
        return torch.logsumexp(elementwise_sum, dim=2)
    
    def transition_backward_sparse(self,
                                   transition:torch.sparse_coo_tensor,
                                   log_emission:torch.Tensor,
                                   log_beta:torch.Tensor):
        """
        Calculate the step backward log probability.
        log beta_t^k = log sum_i exp{log T_t^{k,i} + log e^i(x_{t+1})+log beta_{t+1}^i}

        Parameters
        ----------
        transition : torch.sparse_coo_tensor shape [B,N,N]
            The transition matrix at current time point.
        log_emission : torch.Tensor shape [B,N]
            The log emission probability at time point t+1.
        log_beta : torch.Tensor shape [B,N]
            The t+1 log beta posterior probability.
        normalization : bool, opotional
            If we want to normalize the transition matrix.

        Returns
        -------
        The log beta posterior probabtiliy at current time point t.

        """
        B = log_emission.unsqueeze(dim = 1) + log_beta.unsqueeze(dim = 1)
        result = log_domain_sparse_matmul(transition, B,dim = 2).to_dense()
        result[result == 0] = self.log_epsilon.to(self.device) + B.squeeze(dim = 1)[result == 0]
        return result
    
    def transition_backward_compact(self,
                                    transition:torch.Tensor,
                                    log_emission:torch.Tensor,
                                    log_beta:torch.Tensor,
                                    normalization:bool = False):
        """
        Calculate the step backward log probability.
        log beta_t^k = log sum_i exp{log T_t^{k,i} + log e^i(x_{t+1})+log beta_{t+1}^i}

        Parameters
        ----------
        transition : torch.sparse_coo_tensor shape [B,N,b+1]
            The compact transition matrix at current time point.
        log_emission : torch.Tensor shape [B,N]
            The log emission probability at time point t+1.
        log_beta : torch.Tensor shape [B,N]
            The t+1 log beta posterior probability.
        normalization : bool, opotional
            If we want to normalize the transition matrix.

        Returns
        -------
        The log beta posterior probabtiliy at current time point t.

        """
        if normalization:
            log_transition = normalize(transition,p=1,dim=2).log()
        else:
            log_transition = transition.log()
        B = log_emission + log_beta
        multiplication = B[:,self.target_mapping] #resulting shape [B,N,b+1]
        return torch.logsumexp(log_transition + multiplication,dim = -1)
    
    def transition_prob(self, 
                        transition_matrix:torch.Tensor, 
                        log_alpha:torch.Tensor,
                        normalization:bool = True):
        """
        Calculate the transition probability given the current transition
        matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.Tensor wish shape [B,N,N]
            Current transition matrix, B is the batch size and N is the number
            of states.
        log_alpha : torch.Tensor with shape [B,N]
            Current time alpha matrix. The default is None.
        normalization : bool, optional
            If we want to normalize the transition matrix.
        Returns
        -------
        The log transition probability with shape [B,N].

        """
        if normalization:
            log_transition = normalize(transition_matrix,p=1,dim=2).log()
        else:
            log_transition = transition_matrix.log()
        return log_domain_matmul(log_alpha, log_transition)
    
    def transition_prob_sparse(self, 
                               transition_matrix:torch.sparse_coo_tensor, 
                               log_alpha:torch.Tensor):
        """
        Calculate the transition probability given the current transition
        matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.sparse_coo_tensor wish shape [B,N,N]
            Current transition matrix, B is the batch size and N is the number
            of states.
        log_alpha : torch.Tensor with shape [B,N]
            Current time alpha matrix. The default is None.
        Returns
        -------
        The log transition probability with shape [B,N].

        """
        result = log_domain_sparse_matmul(transition_matrix, log_alpha.unsqueeze(dim = 2), dim = 1).to_dense()
        result[result == 0] = self.log_epsilon.to(self.device) + log_alpha[result == 0]
        return result
    
    def transition_prob_compact(self,
                                transition_matrix:torch.Tensor, 
                                log_alpha:torch.Tensor,
                                normalization:bool = False):
        """
        Calculate the transition probability given the compact transition
        matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.Tensor wish shape [B,N,b+1]
            Current transition matrix, B is the batch size and N is the number
            of states, b is the number of base.
        log_alpha : torch.Tensor with shape [B,N]
            Current time alpha matrix. The default is None.
        normalization : bool, optional
            If we want to normalize the transition matrix.
        Returns
        -------
        The log transition probability with shape [B,N].

        """
        if normalization:
            log_transition = normalize(transition_matrix,p=1,dim=2).log()
        else:
            log_transition = transition_matrix.log()
        indexing = self.source_mapping
        multiplication = log_alpha[:,indexing]
        return torch.logsumexp(log_transition + multiplication,dim = -1)
    
    def transition_max(self,
                        transition_matrix:torch.Tensor, 
                        log_delta:torch.Tensor,
                        normalization:bool = True):
        """
        Calculate the transition probability of the most probable path given 
        the current transition matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.Tensor wish shape [B,N,N]
            Current transition matrix, B is the batch size and N is the number
            of states.
        log_delta : torch.Tensor with shape [B,N]
            Current time alpha matrix.
        normalization : bool, optional
            If we want to normalize the transition matrix.
        Returns
        -------
        (log_prob, max_idx)
        The log transition probability of most probable path and the argmax 
        index.

        """
        if normalization:
            log_transition = normalize(transition_matrix,p=1,dim=2).log()
        else:
            log_transition = transition_matrix.log()
        return torch.max(log_delta.unsqueeze(dim = 2) + log_transition, dim = 1)
    
    def transition_max_compact(self,
                               transition_matrix:torch.Tensor, 
                               log_delta:torch.Tensor,
                               normalization:bool = True):
        """
        Calculate the transition probability of the most probable path given 
        the current transition matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.Tensor wish shape [B,N,N]
            Current transition matrix, B is the batch size and N is the number
            of states.
        log_delta : torch.Tensor with shape [B,N]
            Current time alpha matrix.
        normalization : bool, optional
            If we want to normalize the transition matrix.
        Returns
        -------
        (log_prob, max_idx)
        The log transition probability of most probable path and the argmax 
        index.

        """
        if normalization:
            log_transition = normalize(transition_matrix,p=1,dim=2).log()
        else:
            log_transition = transition_matrix.log()
        indexing = self.source_mapping
        multiplication = log_delta[:,indexing]
        max_prob,max_idx = torch.max(log_transition + multiplication,dim = -1)
        source_mapping_ex = self.source_mapping.expand(max_idx.shape[0],*self.source_mapping.shape)
        max_idx = source_mapping_ex.gather(-1,max_idx.unsqueeze(-1)).squeeze(-1)
        return max_prob,max_idx
    
    def transition_max_sparse(self,
                        transition_matrix:torch.Tensor, 
                        log_delta:torch.Tensor):
        """
        Calculate the transition probability of the most probable path given 
        the current transition matrix and current forward probability.

        Parameters
        ----------
        transition_matrix : torch.Tensor wish shape [B,N,N]
            Current transition matrix, B is the batch size and N is the number
            of states.
        log_delta : torch.Tensor with shape [B,N]
            Current time alpha matrix.
        Returns
        -------
        (log_prob, max_idx)
        The log transition probability of most probable path and the argmax 
        index.

        """
        B,N = log_delta.shape
        result,max_idx = log_domain_sparse_max(transition_matrix, log_delta.unsqueeze(dim = 2), dim = 1)
        max_idx[result == 0] = torch.arange(N,device = self.device).repeat(B,1)[result == 0]
        result[result == 0] = self.log_epsilon.to(self.device) + log_delta[result == 0]
        return result,max_idx
        
    def _to_device(self,data,device):
        if isinstance(data, (list,tuple)):
            return [self._to_device(x,device) for x in data]
        if isinstance(data, (dict)):
            temp_dict = {}
            for key in data.keys():
                temp_dict[key] = self._to_device(data[key],device)
            return temp_dict
        return data.to(device, non_blocking=True)
    
    def _get_device(self,device):
        if device is None:
            if torch.cuda.is_available():
                return torch.device('cuda')
            else:
                return torch.device('cpu')
        else:
            return torch.device(device)
    
    
    
    def maximization(self, observation:torch.Tensor, 
                     marginal_probability:torch.Tensor,
                     update = False,
                     lr = 0.1,
                     weight_decay = 0.0,
                     momentum = 0.9):
        self.emission._accumulate_sufficient_statistics(observation,
                                                        marginal_probability)
        if update:
            return self.emission.update_parameters(lr = lr,
                                                   weight_decay = weight_decay,
                                                   momentum = momentum)
    
    def to_dense(self, sparse_matrix_array:Union[np.array,torch.sparse_coo_tensor]):
        """
        Make an array of sparse matrix into a dense 3D tensor.
    
        Parameters
        ----------
        sparse_matrix_array : Union[np.array,torch.sparse_coo_tensor]
            A 1D np array, each element is a torch sparse tensor or a sparse torch
            coo tensor with shape [B,N,N]
        epsilon : float
            A small number to add to diagonal to make numerical stability.
    
        Returns
        -------
        TYPE
            DESCRIPTION.
    
        """
        n_states = sparse_matrix_array[0].size()[0]
        if type(sparse_matrix_array) == np.ndarray:
            dense_array = torch.stack([x.to_dense() for x in sparse_matrix_array])
            if dense_array.device != self.device:
                dense_array = dense_array.to(self.device)
            return dense_array + self.epsilon * torch.eye(n_states,device = self.device)
        else:
            return sparse_matrix_array.to_dense() + (self.epsilon * torch.eye(n_states,device = self.device))

def log_domain_matmul(log_A, log_B):
    """
    log_A : m x n
    log_B : m x n x p or n x p
    output : m x p matrix
    
    Normally, a matrix multiplication
    computes out_{i,j} = sum_k A_{i,k} x B_{i,k,j}

	A log domain matrix multiplication
	computes out_{i,j} = logsumexp_k log_A_{i,k} + log_B_{i,k,j}
	"""
    dim_B = len(log_B.shape)
    log_A = log_A.unsqueeze(dim = 2)
    if dim_B == 2:
        log_B = log_B.unsqueeze(dim = 0)
    elementwise_sum = log_A + log_B
    out = torch.logsumexp(elementwise_sum, dim=1)
    return out

def log_domain_matmul_mapping(log_A, log_B, index_map):
    """
    log_A : m x n
    log_B : m x p x b or p x b
    index_map : p x b
    output : m x p matrix

	The following log domain matrix multiplication with index mapping is calculated
	computes out_{i,j} = logsumexp_k log_A_{i,k} + log_B_{i,k,j}
	"""
    dim_B = len(log_B.shape)
    log_A = log_A.unsqueeze(dim = 2)
    if dim_B == 2:
        log_B = log_B.unsqueeze(dim = 0)
    elementwise_sum = log_A + log_B
    out = torch.logsumexp(elementwise_sum, dim=1)
    return out

def density(sparse_tensor:torch.sparse_coo_tensor):
    return len(sparse_tensor.indices())/torch.numel(sparse_tensor)

def log_domain_sparse_matmul(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    idxs,update = log_domain_sparse_product(A,log_B)
    n_dims_A = idxs.shape[0]
    shape_A = torch.tensor(A.shape)
    remain_dims = np.arange(n_dims_A)
    remain_dims = np.delete(remain_dims,dim)
    remain_idxs = idxs[remain_dims,:].T.tolist()
    update = update.tolist()
    key_func = lambda x: x[1]
    update = sorted(zip(update,remain_idxs),key = key_func)
    nested = [ (k,list(g)) for k,g in itertools.groupby(update,key = key_func)]
    nested_vals = [[y[0] for y in x[1]] for x in nested]
    nested_idxs = [x[0] for x in nested]
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    return torch.sparse_coo_tensor(indices = list(zip(*nested_idxs)),
                                   values = torch.logsumexp(padded,dim = 1),
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device)

def log_domain_sparse_max(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    idxs,update = log_domain_sparse_product(A,log_B)
    n_dims_A = idxs.shape[0]
    shape_A = torch.tensor(A.shape)
    remain_dims =np.delete(np.arange(n_dims_A),dim)
    remain_idxs = list(zip(*[idxs[x].tolist() for x in remain_dims]))
    reduced_idxs = idxs[dim].tolist()
    update = update.tolist()
    key_func = lambda x: x[2]
    update = sorted(zip(update,reduced_idxs,remain_idxs),key = key_func)
    nested = [ (k,list(g)) for k,g in itertools.groupby(update,key = key_func)]
    nested_vals = [[y[0] for y in x[1]] for x in nested]
    nested_idxs = list(zip(*[x[0] for x in nested]))
    nested_reduced_idxs = [[y[1] for y in x[1]] for x in nested]
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    padded_reduced_idxs = torch.tensor([x + [-1]*(max_cols - len(x)) for x in nested_reduced_idxs],device = A.device,dtype = torch.long)
    result,argmax = torch.max(padded,dim = 1)
    argmax = torch.gather(padded_reduced_idxs,dim = 1,index = argmax[:,None]).squeeze(dim = 1)
    max_idx = torch.zeros(shape_A[remain_dims].tolist(),device = A.device,dtype = torch.long)
    max_idx[tuple(nested_idxs)] = argmax
    return torch.sparse_coo_tensor(indices = nested_idxs,
                                   values = result,
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device).to_dense(),max_idx


def log_domain_sparse_product(A:torch.sparse_coo_tensor, log_B:torch.Tensor):
    """
    Do a sparse-dense tensor production.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.

    Returns
    -------
    idxs,vals

    """
    A = A.coalesce()
    idxs = A.indices()
    log_vals = A.values().log()
    shape_A = torch.tensor(A.shape)
    shape_B = torch.tensor(log_B.shape)
    n_dims_A = idxs.shape[0]
    n_dims_B = len(shape_B)
    assert n_dims_A == n_dims_B, "Tensor has different number of dimensions."
    assert torch.all(torch.logical_or(shape_A == shape_B,shape_B == 1)), "Shape mismatch, got {} and {}. Broadcast only supported on dense tensor.".format(shape_A,shape_B)
    idxs_B = idxs.clone()
    idxs_B[torch.where(shape_B==1)] = 0
    update = log_B[tuple(idxs_B)] + log_vals
    return idxs,update
    
    
if __name__ == "__main__":
    # Module test
    from xron.nrhmm.hmm_input import Kmer_Dataset
    from torch.utils.data.dataloader import DataLoader
    from xron.xron_train_base import DeviceDataLoader
    from matplotlib import pyplot as plt
    import time
    ## Construct the simulation data
    np.random.seed(2022)
    device = "cuda"
    n_states = 10
    background = 0.1
    T_max = 20000
    n_samples = 1500
    batch_size = 50
    single_length = 500
    segment_length = 20
    epoches_n = 100
    means = np.random.rand(n_states) * 5
    init_means = np.random.rand(n_states)
    covs = np.random.rand(n_states)
    init_covs = np.asarray([2]*n_states)
    z_true = np.asarray([[x]*segment_length for x in np.random.randint(0,n_states,T_max//segment_length)]).flatten()
    x = means[z_true] + covs[z_true]*(np.random.rand(T_max)-0.5)*2
    
    transitions = np.empty(T_max,dtype = object)
    for t in np.arange(T_max-1):
        source = [z_true[max(t-2*segment_length,0)],z_true[max(t-2*segment_length,0)],z_true[max(t-segment_length,0)],z_true[max(t-segment_length,0)],z_true[t],z_true[t],z_true[min(t+segment_length,T_max-2)],z_true[min(t+segment_length,T_max-2)],z_true[min(t+2*segment_length,T_max-2)]]
        target = [z_true[max(t-2*segment_length,0)],z_true[max(t-segment_length,0)],z_true[max(t-segment_length,0)],z_true[t],z_true[t],z_true[min(t+segment_length,T_max-2)],z_true[min(t+segment_length,T_max-2)],z_true[min(t+2*segment_length,T_max-2)],z_true[min(t+2*segment_length,T_max-2)]]
        source,target = list(zip(*set(list(zip(source,target)))))
        curr_transition = torch.sparse_coo_tensor([source,target], [1.0]*len(source),(n_states,n_states))
        transitions[t] = curr_transition
    transitions[T_max-1] = torch.sparse_coo_tensor([[z_true[-1]],[z_true[-1]]],[1],(n_states,n_states))
    def sampling(x, n_samples,length = single_length):
        T_max = len(x)
        start = np.random.randint(low = 0,high = T_max-length, size = n_samples)
        idxs = np.asarray([np.arange(x,x+length) for x in start])
        return idxs, x[idxs]
    idxs,xs = sampling(x,n_samples)
    xs = xs.astype(np.float32)
    Transitions = transitions[idxs].T
    list_transitions = [list(x) for x in list(Transitions.T)]
    labels = z_true[idxs]
    starts = np.zeros((n_samples,n_states)) + background
    starts[range(n_samples),z_true[idxs[:,0]]] = 1.
    Ls = np.asarray([single_length]*n_samples)

    # Construct the model and training using the direct gradient descent method.
    # emission = GaussianEmissions(init_means[:,None], init_covs[:,None])
    # hmm = RHMM(emission,normalize_transition=False,sparse_operation=False)
    # optimizer = torch.optim.Adam(hmm.parameters(),lr = 1e-1)
    # perm = np.arange(n_samples)
    # update_every_n = 5
    # dataset = Kmer_Dataset(xs, Ls, list_transitions)
    # loader = DataLoader(dataset,batch_size = batch_size, shuffle = False)
    # loader = DeviceDataLoader(loader,device = device)
    # for epoch_i in np.arange(epoches_n):
    #     for i,batch in enumerate(loader):
    #         batch_x = batch['signal']
    #         transition = batch['labels']
    #         start = torch.zeros((batch_size,n_states)).to(device)
    #         batch_label = labels[(i)*batch_size%n_samples:(((i+1)*batch_size-1)%n_samples+1)]
    #         batch_duration = batch['duration']
    #         log_prob = hmm.forward(batch_x, batch_duration, transition)
    #         loss = -log_prob.mean()
    #         optimizer.zero_grad()
    #         loss.backward()
    #         optimizer.step()
    #         hmm.emission.clamp_cov()
    #         error = np.linalg.norm(emission.means.cpu().squeeze(dim = 1).detach().numpy() - means,ord = 1)/n_states
            
    #         path,logit = hmm.viterbi_decode(batch_x, batch_duration, transition)
    #         mask = np.arange(single_length)[None,:].repeat(batch_x.shape[0],axis = 0)<batch_duration[:,None].detach().cpu().numpy()
    #         accr = (batch_label[mask] == path.detach().cpu().numpy()[mask]).sum()/(sum(batch_duration.detach().cpu().numpy()))
    #         print("%d epoch-%d batch: loss %.2f, error %.2f, accuracy %.2f"%(epoch_i,i,loss.cpu().detach().numpy(),error,accr))

    # Train using Expectation-Maximization algorithm
    visual = False
    emission = GaussianEmissions(init_means[:,None], init_covs[:,None])
    hmm = RHMM(emission,normalize_transition=False)
    optimizer = torch.optim.Adam([hmm.emission.means],lr = 1)
    perm = np.arange(n_samples)
    update_every_n = 1
    dataset = Kmer_Dataset(xs, Ls, list_transitions,labels = labels)
    loader = DataLoader(dataset,batch_size = batch_size, shuffle = True)
    loader = DeviceDataLoader(loader,device = device)
    for epoch_i in np.arange(epoches_n):
        for i,batch in enumerate(loader):
            start_time = time.time()
            batch_x = batch['signal']
            transition = batch['labels']
            batch_label = batch['y']
            start = torch.zeros((batch_size,n_states)).to(device)
            batch_duration = batch['duration']
            with torch.no_grad():
                s1 = time.time()
                gamma = hmm.expectation(batch_x, batch_duration, transition)
                print("Expectation time %.2f"%(time.time() - s1))
                g = hmm.maximization(batch_x,gamma,update = (i%update_every_n==0),momentum = 0.9,lr = 1. ,weight_decay = 0.0)
            if i%update_every_n == 0:
                error = np.linalg.norm(emission.means.cpu().squeeze(dim = 1).detach().numpy() - means,ord = 1)/n_states
                log_prob = hmm.forward(batch_x, batch_duration, transition)
                loss = -torch.logsumexp(log_prob, dim = 1).mean()
                path,logit = hmm.viterbi_decode(batch_x, batch_duration, transition)
                mask = np.arange(single_length)[None,:].repeat(batch_size,axis = 0)<batch_duration[:,None].detach().cpu().numpy()
                accr = (batch_label[mask].detach().cpu().numpy() == path.detach().cpu().numpy()[mask]).sum()/(sum(batch_duration.detach().cpu().numpy()))
                print("Epoch %d - Batch %d: loss %.2f, error %.2f, accuracy %.2f, time per batch %.2f"%(epoch_i,i,loss,error,accr,time.time()-start_time))
                max_variance = np.max(hmm.emission.certainty.cpu().numpy())
                print("Rescale the covariance to %.2f"%(max_variance))
                hmm.emission.cov[:] = 2*max_variance
                # print("Epoch %d - Batch %d: error %.2f, accuracy %.2f"%(epoch_i,i,error,accr))
                if visual:
                    plt.plot(logit[0,np.arange(single_length),batch_label[0]].detach().cpu().numpy(),label = "Oirignal.")
                    plt.plot(logit[0,np.arange(single_length),path[0]].detach().cpu().numpy(),label = "Realigned.")
                    plt.ylabel("Log probability.")
                    plt.xlabel("Time")
                    plt.figure()
                    plt.plot(batch_x[0].cpu().numpy())
                    rc_sig = [hmm.emission.means[x].item() for x in path[0]]
                    plt.plot(rc_sig)
                    


"""
Created on Tue Aug  3 23:24:39 2021

@author: Haotian Teng
"""
import numpy as np
import os
import sys
import argparse


def main(args):
    chunks = []
    counts = []
    for i in args.input:
        try:
            c = np.load(i)
        except FileNotFoundError:
            print("Didn't find %s, skip."%(i))
            continue
        counts.append(len(c))
        chunks.append(c)
        print("Read %d chunks from %s"%(len(c),i))
    if args.equal_size:
        min_size = min(counts)
        chunks = [x[:min_size] for x in chunks]
    chunk_all = np.concatenate(chunks,axis = 0)
    np.save(os.path.join(args.output,'chunks_all.npy'),chunk_all)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Merge signal chunks from different folder.')
    parser.add_argument('-i', '--input', default = None,
                        help = "The input signal chunks, separate by comma.")
    parser.add_argument('-o', '--output', required = True,
                        help = "The output folder of the merged dataset.")
    parser.add_argument('--equal_size',default = False, type = bool,
                        help = "If make the size from each chunks equal.")
    args = parser.parse_args(sys.argv[1:])
    args.input = args.input.split(',')
    if not os.path.exists(args.output):
        os.makedirs(args.output)
    main(args)


"""
@author: Haotian Teng
"""
import os 
import sys
import torch
import argparse
import numpy as np
from typing import Union,List
from itertools import chain
from torchvision import transforms
import torch.utils.data as data
from torch.utils.data.dataloader import DataLoader
from xron.xron_input import Dataset, ToTensor, NumIndex,rna_filt,dna_filt
from xron.xron_train_base import Trainer, DeviceDataLoader, load_config
from xron.xron_model import REVCNN,DECODER_CONFIG,CRNN,CRITIC_CONFIG,CRITIC,MM_CONFIG,MM
from xron.xron_label import MetricAligner
from torch.distributions.one_hot_categorical import OneHotCategorical as OHC

class VAETrainer(Trainer):
    def __init__(self,
                 train_dataloader:DataLoader,
                 encoder:CRNN,
                 critic:CRITIC,
                 decoder:REVCNN,
                 mm:MM,
                 config:Union[DECODER_CONFIG,MM_CONFIG],
                 aligner:MetricAligner,
                 device:str = None,
                 eval_dataloader:DataLoader = None):
        """

        Parameters
        ----------
        train_dataloader : DataLoader
            Training dataloader.
        encoder: CRNN
            A Convolutional-Recurrent Neural Network
        critic: CRNN
            A critic network with the same structure as the encoder.
        decoder : REVCNN
            REVCNN decoder
        mm: MM
            Markov Model instance.
        device: str
            The device used to train the model, can be 'cpu' or 'cuda'.
            Default is None, use cuda device if it's available.
        config: Union[DECODER_CONFIG,MM_CONFIG]
            A CONFIG class contains unsupervised training configurations. Need 
            to contain at least these parameters: keep_record, device and 
            grad_norm.
        eval_dataloader : DataLoader, optional
            Evaluation dataloader, if None training dataloader will be used.
            The default is None.

        """
        super().__init__(train_dataloader=train_dataloader,
                         nets = {"encoder":encoder,
                                 "critic":critic,
                                 "decoder":decoder,
                                 "mm":mm},
                         config = config,
                         device = device,
                         eval_dataloader = eval_dataloader)
        self.encoder = encoder
        self.critic = critic
        self.decoder = decoder
        self.mm = mm
        self.aligner = aligner
        self.train_config = config.TRAIN
        self.decay = self.train_config["decay"]
        self.global_step = 0
        self.awake = False
        self.score_average = 0
        self.records = {'rc_losses':[],
                        'rc_different':[],
                        'entropy_losses':[],
                        'alignment_score':[],
                        'valid_alignment':[]}
        
    def train(self,
              epoches:int,
              optimizers:List[torch.optim.Optimizer],
              save_cycle:int,
              save_folder:str):
        """
        Train the encoder-decodr nets.

        Parameters
        ----------
        epoches : int
            Number of epoches to train.
        optimizers : List[torch.optim.Optimizer]
            A list of two optimizers, the first one is optimizer training the
            encoder parameters and the second one for decoder parameters.
        save_cycle : int
            Save every save_cycle batches.
        save_folder : str
            The folder to save the model and training records.

        Returns
        -------
        None.

        """
        self.save_folder = save_folder
        self._save_config()
        records = self.records
        preheat = self.train_config["preheat"]
        e_decay = self.train_config["epsilon_decay"]
        epsilon = self.train_config["epsilon"]
        alpha = self.train_config["alpha"]
        beta = self.train_config["beta"]
        gamma = self.train_config["gamma"]
        opt_e,opt_c,opt_g,opt_d = optimizers
        for epoch_i in range(epoches):
            for i_batch, batch in enumerate(self.train_ds):
                if self.global_step < preheat:
                    self.epsilon = 0
                else:
                    self.epsilon = max(1+(self.global_step - preheat)*e_decay,epsilon)
                losses = self.train_step(batch)
                opt_c.zero_grad()
                losses[-1].backward()
                opt_c.step()
                losses = self.train_step(batch)
                opt_g.zero_grad()
                losses[-2].backward()
                opt_g.step()
                losses = self.train_step(batch)
                opt_d.zero_grad()
                losses[-3].backward()
                opt_d.step()
                if self.global_step > preheat:
                    losses = self.train_step(batch)
                    loss = -alpha*losses[0]+beta*losses[1]+gamma*losses[2] #Maximize -H(q), minimize -(cross entropy loss)
                    opt_e.zero_grad()
                    loss.backward()
                    opt_e.step()
                if (self.global_step+1)%save_cycle==0:
                    self.save()
                    self.mm.save_embedding(alphabeta=self.config.CTC['alphabeta'])
                    eval_i,valid_batch = next(enumerate(self.eval_ds))
                    valid_error,valid_perm = self.valid_step(valid_batch)
                    records['entropy_losses'].append(losses[0].detach().cpu().numpy()[()])
                    records['valid_alignment'].append(valid_error)
                    print("Epoch %d Batch %d, entropy_loss %f, rc_loss %f, encoded valid_error %f, perm:%s, alignment_loss %f"%(epoch_i, i_batch, losses[0], losses[3], valid_error, valid_perm, losses[2]))
                    records['alignment_score'].append(losses[2].detach().cpu().numpy()[()])
                    records['rc_different'].append(losses[1].detach().cpu().numpy()[()])
                    records['rc_losses'].append(losses[3].detach().cpu().numpy()[()])
                    self._update_records()
                losses = None
                torch.nn.utils.clip_grad_norm_(self.parameters, 
                                               max_norm=self.grad_norm)
                self.global_step +=1
        
    def train_step(self,batch):
        encoder = self.encoder
        critic = self.critic
        decoder = self.decoder
        mm = self.mm
        signal = batch['signal']
        logprob = encoder.forward(signal) #[L,N,C]
        mse_predict = critic.forward(signal).permute(1,2,0) #[L,N,C] -> [N,C,L]
        if np.random.rand()<self.epsilon:
            m = OHC(logits = logprob)
            sampling = m.sample().permute([1,2,0]) #[L,N,C]->[N,C,L]
        else:
            sampling = torch.argmax(logprob,dim = 2)
            sampling = torch.nn.functional.one_hot(sampling,num_classes = logprob.shape[2]).permute([1,2,0]).float()
        rc_mm = mm.forward(sampling,device = self.device).permute([0,2,1]) #[N,L,C] -> [N,C,L]
        rc_revcnn = decoder.forward(sampling).permute([0,2,1]) #[N,L,C] -> [N,C,L]
        rc_signal = rc_mm+rc_revcnn
        if self.config.TRAIN['diff_signal']:
            pass
            #TODO implement the encoder to reconstruct diffrential signal.
        mse_combine = decoder.mse_loss(rc_signal,signal)
        mse_mm = decoder.mse_loss(rc_mm,signal)
        sample_logprob = torch.sum(logprob.permute([1,2,0])*sampling,axis = (1,2))
        entropy_loss = torch.mean(sample_logprob**2)
        raw_seq = torch.argmax(sampling,dim = 1)
        critic_loss = torch.mean(critic.mse_loss(mse_combine.detach(),mse_predict))

        rc_loss = torch.mean(torch.mean(mse_combine - mse_predict.detach(),axis = (1,2))*sample_logprob)
        decoder_loss_combine = torch.mean(mse_combine)
        decoder_loss_mm = torch.mean(mse_mm)

        ### Train with aligner
        # identities,perm = self.aligner.permute_align(raw_seq.cpu().detach().numpy())
        # score = np.mean(identities,axis = 1)
        # best_perm = np.argmax(score)
        # score = torch.from_numpy(identities[best_perm]-self.score_average).to(self.device)
        # self.score_average  = self.score_average * self.decay + (1-self.decay)* np.mean(identities[best_perm])
        # return entropy_loss,rc_loss, torch.mean(-score*sample_logprob)
        return entropy_loss, rc_loss, torch.zeros(1).to(self.device),decoder_loss_combine,decoder_loss_mm,critic_loss #Hack the alignment loss out
    
    def valid_step(self,batch):
        net = self.encoder
        signal_batch = batch['signal']
        out = net.forward(signal_batch)
        if 'seq' in batch.keys():
            seq = batch['seq']
            seq_len = batch['seq_len'].view(-1)
            errors,perms = net.permute_error(out,
                           seq,
                           seq_len,
                           alphabet = 'N' + self.config.CTC['alphabeta'],
                           beam_size = self.config.CTC['beam_size'],
                           beam_cut_threshold = self.config.CTC['beam_cut_threshold'])
            best_perm = np.argmin(errors)
            return errors[best_perm],perms[best_perm]
        else:
            out_seq = torch.argmax(out,dim = 2)
            identities,perms = self.aligner.permute_align(out_seq.cpu().detach().numpy())
            score = np.mean(identities,axis = 1)
            best_perm = np.argmax(score)
            return 1-score[best_perm],perms[best_perm]
    
def main(args):
    class CTC_CONFIG(MM_CONFIG):
        CTC = {"beam_size":5,
               "beam_cut_threshold":0.05,
               "alphabeta": "ACGTM",
               "mode":"rna"}
    class TRAIN_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "epsilon":0.1,
                 "epsilon_decay":0,
                 "alpha":0.01, #Entropy loss scale factor
                 "beta": 1., #Reconstruction loss scale factor
                 "gamma":0, #Alignment loss scale factor
                 "preheat":5000,
                 "keep_record":5,
                 "decay":args.decay,
                 "diff_signal":args.diff}
        
    config = TRAIN_CONFIG()
    config.PORE_MODEL["N_BASE"] = len(config.CTC["alphabeta"])
    critic_config = CRITIC_CONFIG()
    print("Read chunks and sequence.")
    chunks = np.load(args.chunks,allow_pickle = True)
    print(chunks.shape)
    reference = np.load(args.seq) if args.seq else None
    ref_len = np.load(args.seq_len) if args.seq_len else None
    print("Construct and load the model.")
    model_f = args.model_folder
    if reference is not None and (reference[0].dtype.kind in ['U','S']):
        if config.CTC['mode'] == 'rna':
            chunks,reference,ref_len = rna_filt(chunks,reference,ref_len)
        elif config.CTC['mode'] == 'dna':
            chunks,reference,ref_len = dna_filt(chunks,reference,ref_len)
        alphabet_dict = {x:i+1 for i,x in enumerate(TRAIN_CONFIG.CTC['alphabeta'])}
        dataset = Dataset(chunks,seq = reference,seq_len = ref_len,transform = transforms.Compose([NumIndex(alphabet_dict),ToTensor()]))
    else:
        dataset = Dataset(chunks,seq = reference,seq_len = ref_len,transform = transforms.Compose([ToTensor()]))
    loader = data.DataLoader(dataset,batch_size = args.batch_size,shuffle = True, num_workers = 4)
    DEVICE = args.device
    loader = DeviceDataLoader(loader,device = DEVICE)
    if args.pretrain_encoder:
        encoder_config = load_config(os.path.join(args.pretrain_encoder,"config.toml"))
        config.CNN,config.RNN,config.FNN = encoder_config.CNN,encoder_config.RNN,encoder_config.FNN
    if args.retrain:
        config_old = load_config(os.path.join(model_f,"config.toml"))
        config_old.TRAIN = config.TRAIN #Overwrite training config.
        config = config_old
    encoder = CRNN(config)
    critic = CRITIC(critic_config)
    decoder = REVCNN(config)
    mm = MM(config)
    aligner = MetricAligner(args.reference)
    t = VAETrainer(loader,encoder,critic,decoder,mm,config,aligner)
    if args.pretrain_encoder:
        t.load(args.pretrain_encoder,update_global_step = False)
    if args.retrain:
        t.load(model_f)
    lr = args.lr
    epoches = args.epoches
    opt_e = torch.optim.Adam(t.encoder.parameters(),lr = lr)
    opt_c = torch.optim.Adam(t.critic.parameters(),lr = lr)
    opt_mm = torch.optim.SGD(t.mm.parameters(),lr = lr)
    opt_revcnn = torch.optim.Adam(t.decoder.parameters(),lr = lr)
    COUNT_CYCLE = args.report
    print("Begin training the model.")
    t.train(epoches,[opt_e,opt_c,opt_mm,opt_revcnn],COUNT_CYCLE,model_f)
        
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training model with chunks and sequence file')
    parser.add_argument('-i', '--chunks', required = True,
                        help = "The .npy file contain chunks.")
    parser.add_argument('-o', '--model_folder', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument('--pretrain_encoder',default = None,
                        help = "The folder contain the pretrained encoder model.")
    parser.add_argument('-r', '--reference', required = True,
                        help = "The reference fastq file.")
    parser.add_argument('--seq', default = None,
                        help="The .npy file contain the sequence.")
    parser.add_argument('--seq_len', default = None,
                        help="The .npy file contain the sueqnece length.")
    parser.add_argument('--device', default = 'cuda',
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--lr', default = 4e-3, type = float,
                        help="Initial learning rate.")
    parser.add_argument('--batch_size', default = 200, type = int,
                        help="Training batch size.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument('--report', default = 10, type = int,
                        help = "The interval of training rounds to report.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--decay', type = float, default = 0.99,
                        help="The decay factor of the moving average.")
    parser.add_argument('--diffSig',action="store_true",dest = "diff",
                        help="If the input chunks are diffrential signal.")
    args = parser.parse_args(sys.argv[1:])
    if not os.path.isdir(args.model_folder):
        os.mkdir(args.model_folder)
    main(args)



# Copyright 2017 The Chiron Authors. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Created on Tue May  2 15:39:29 2017

from __future__ import absolute_import
from __future__ import print_function
import difflib
import math
import operator
import time
from xron.utils.seq_op import list2string,string2list
from collections import Counter
from itertools import groupby
try:
    from itertools import zip_longest as zip_longest
except:
    from itertools import izip_longest as zip_longest
import numpy as np
import six
from six.moves import range
from Bio import pairwise2

def mapping(full_path, blank_pos=4):
    """Perform a many to one mapping in the CTC paper, merge the repeat and remove the blank
    Input:
        full_path:a vector of path, e.g. [1,0,3,2,2,3]
        blank_pos:The number regarded as blank"""
    full_path = np.asarray(full_path)
    merge_repeated = np.asarray([k for k, g in groupby(full_path)])
    blank_index = np.argwhere(merge_repeated == blank_pos)
    return np.delete(merge_repeated, blank_index)

def group_consecutives(vector, step=1):
    group = list()
    group_list = list()
    expect = None
    for x in vector:
        if (x != expect) and (expect is not None):
            group_list.append(group)
            group = []
        group.append(x)
        expect = x + step
    group_list.append(group)
    return group_list


###########################Section decoding method#############################
def section_decoding(logits, blank_thres=0.6, base_type=0):
    """Implemented the decoding method described in ftp://ftp.idsia.ch/pub/juergen/icml2006.pdf
    Find the best path between the section that divided by blank logits < 0.9
    
    logits: [batch_size,seg_length,neucloe_type+1]
    base_type: 0:dna 1:methylation 2:rna
    """
    prob = np.exp(logits) / (np.sum(np.exp(logits), axis=2)[:, :, None])
    if base_type == 0:
        blank_pos = 4
    mask = prob[:, :, blank_pos] < blank_thres
    batch_size, seg_len, nc_type = prob.shape
    bpreads = list()
    bpread = list()
    for batch_i in range(batch_size):

        group_list = group_consecutives(np.where(mask[batch_i, :])[0])

        bpread = []
        for group in group_list:
            if len(group) == 0:
                continue
            bpread.append(4)
            #            most_prob_path = best_path(prob[batch_i,group,:],base_type = base_type)
            most_mc_path = mc_path(logits[batch_i, group, :],
                                   base_type=base_type)
            most_prob_path = string2list(most_mc_path[0])
            bpread += most_prob_path
        bpreads.append(list2string(mapping(bpread), base_type=base_type))
    return (bpreads)


def best_path(logits, base_type):
    """Enumerate decoder,*slow*"""
    T, base_num = logits.shape
    accum_prob = {}
    for i in range(base_num ** T):
        prob = 1
        index_list = []
        for j in range(T):
            index = i / base_num ** j % base_num
            prob *= logits[j, index]
            index_list.append(index)
        index_list = mapping(index_list)
        if len(index_list) > 0:
            key = list2string(index_list, base_type=base_type)
            accum_prob.setdefault(key, 0)
            accum_prob[key] += prob
    most_prob_path = max(six.iteritems(accum_prob), key=operator.itemgetter(1))[
        0]
    return string2list(most_prob_path, base_type=base_type)


def mc_path(logits, base_type, sample_n=300):
    """Manto Carlo decoder
    Input Args:
        logits:[T,base_num]
        base_tyep: 0:normal dna+blank
        sample_n: Times of sample used in the Manto Carlo simulation.
    """
    logits_shape = logits.shape
    bpreads = list()
    qc_score = list()
    prob = np.exp(logits) / (np.sum(np.exp(logits), axis=2)[:, :, None])
    base_num = logits_shape[-1]
    T = logits_shape[-2]
    interval = np.zeros((T, base_num))
    interval[:, 0] = prob[:, 0]
    for i in range(1, base_num - 1):
        interval[:, i] = interval[:, i - 1] + prob[:, i]
    interval[:, 4] = 1

    sample_index = np.zeros((sample_n, T))
    sample = np.random.random((sample_n, T))
    for j in range(T):
        sample_index[:, j] = np.searchsorted(interval[i, j, :], sample[:, j],
                                             side='left')
    merge_path = list()
    for repeat_i in range(sample_n):
        ###This step may be slow, considering implemented in C
        temp_path = mapping(sample_index[repeat_i, :])
        ###
        merge_path.append(list2string(temp_path, base_type=base_type))
    path_count = Counter(merge_path)
    print(path_count)
    max2path = path_count.most_common(2)
    p1 = max2path[0][1] / float(sample_n)
    p2 = max2path[1][1] / float(sample_n)
    qc_score.append(10 * math.log10(p1 / p2))
    bpreads.append(max2path[0][0])
    return bpreads


def mc_decoding(logits, base_type, sample_n=300):
    """Manto Carlo decoder
    Input Args:
        logits:[batch_size,T,base_num] or [T,base_num]
        base_tyep: 0:normal dna+blank
        sample_n: Times of sample used in the Manto Carlo simulation.
    """
    logits_shape = logits.shape
    if len(logits_shape) == 2:
        logits = [logits]
        batch_size = 1
    else:
        batch_size = logits_shape[0]
    bpreads = list()
    qc_score = list()
    prob = np.exp(logits) / (np.sum(np.exp(logits), axis=2)[:, :, None])
    base_num = logits_shape[-1]
    T = logits_shape[-2]
    interval = np.zeros((batch_size, T, base_num))
    interval[:, :, 0] = prob[:, :, 0]
    for i in range(1, base_num - 1):
        interval[:, :, i] = interval[:, :, i - 1] + prob[:, :, i]
    interval[:, :, 4] = 1

    sample_index = np.zeros((sample_n, T))
    for i in range(batch_size):
        print(i)
        sample = np.random.random((sample_n, T))
        for j in range(T):
            sample_index[:, j] = np.searchsorted(interval[i, j, :],
                                                 sample[:, j], side='left')
        merge_path = list()
        for repeat_i in range(sample_n):
            ###This step may be slow, considering implemented in C
            temp_path = mapping(sample_index[repeat_i, :])
            ###
            merge_path.append(list2string(temp_path, base_type=base_type))
        path_count = Counter(merge_path)
        print(path_count)
        max2path = path_count.most_common(2)
        p1 = max2path[0][1] / float(sample_n)
        p2 = max2path[1][1] / float(sample_n)
        qc_score.append(10 * math.log10(p1 / p2))
        bpreads.append(max2path[0][0])
    return bpreads


###############################################################################

#########################Simple assembly method################################
def simple_assembly_kernal(bpread, prev_bpread,error_rate, jump_step_ratio):
    """
    Kernal function of the assembly method.
    log_P ~ x*log((N*n1/L)) - log(x!) + Ns * log(P1/0.25) + Nd * log(P2/0.25)
    bpread: current read.
    prev_bpread: previous read.
    error_rate: Average basecalling error rate.
    jump_step_ratio: Jump step/Segment len
    """
    back_ratio = 6.5 * 10e-4
    p_same = 1 - 2*error_rate + 26/25*(error_rate**2)
    p_diff = 1 - p_same
    ns = dict() # number of same base
    nd = dict()
    log_px = dict()
    N = len(bpread)
    match_blocks = difflib.SequenceMatcher(a=bpread,b=prev_bpread).get_matching_blocks()
    for idx, block in enumerate(match_blocks):
        offset = block[1] - block[0]
        if offset in ns.keys():
            ns[offset] = ns[offset] + match_blocks[idx][2]
        else:
            ns[offset] = match_blocks[idx][2]
        nd[offset] = 0
#    for offset in range(-3,len(prev_bpread)):
#        pair = zip_longest(prev_bpread[offset:],bpread[:-offset],fillvalue=None)
#        comparison = [int(i==j) for i,j in pair]
#        ns[offset] = sum(comparison)
#        nd[offset] = len(comparison) - ns[offset]
    for key in ns.keys():
        if key < 0:
            k = -key
            log_px[key] = k*np.log((back_ratio)*N*jump_step_ratio) - sum([np.log(x+1) for x in range(k)]) +\
            ns[key]*np.log(p_same/0.25) + nd[key]*np.log(p_diff/0.25)
        else:
            log_px[key] = key*np.log(N*jump_step_ratio) - sum([np.log(x+1) for x in range(key)]) +\
            ns[key]*np.log(p_same/0.25) + nd[key]*np.log(p_diff/0.25)
    disp = max(log_px.keys(),key = lambda x: log_px[x])
    return disp,log_px[disp]

def global_alignment_kernal(bpread, prev_bpread):
    gap_open = -5
    gap_extend = -2
    mismatch = -3
    match = 1
    min_block_size = 3
    global_alignment = pairwise2.align.globalms(prev_bpread,bpread,match,mismatch,gap_open,gap_extend)
    if len(global_alignment) == 0:
        print(bpread)
        print(prev_bpread)
        raise ValueError("Alignment not found")
    blocks = match_blocks(global_alignment[0])
#    if criteria == 'first':
#        for block in blocks:
#            if block[0] >= min_block_size:
#                disp = block[1] - block[2]
#                break
#    elif criteria == "max":
    block = max(blocks, key = lambda x: x[0])
    disp = block[1] - block[2]
    if disp is None:
        disp = blocks[0][1] - blocks[0][2]
    return disp

def glue_kernal(bpread,prev_bpread):
    """
    This is a alignment for a larger jump step.
    A good setting would be jumpstep ~ 0.95 * segment_len
    """
    prev_n = len(prev_bpread)
    n = len(bpread)    
    max_overlap = min(math.floor(0.1 * prev_n),n)
    max_hit_disp = (0,0)
    for i in range(1,max_overlap):
        head=bpread[:i]
        tail=prev_bpread[-i:]
        head = np.asarray([x for x in head])
        tail = np.asarray([x for x in tail])
        score = 2*sum(head==tail) - i
        if score > max_hit_disp[1]:
            max_hit_disp = (i,score)
    disp = prev_n - max_hit_disp[0]
    return disp

def stick_kernal(bpread,prev_bpread):
    """
    Stick assembly,so basically it's just patch the reads together.
    """
    return(len(prev_bpread))

def simple_assembly(bpreads, jump_step_ratio, error_rate = 0.2,kernal = 'global',alphabeta = 'ACGT'):
    """
    Assemble the read from the chunks. Log probability is 
    Args:
        bpreads: Input chunks.
        jump_step_ratio: Jump step divided by segment length.
        error_rate: An estimating basecalling error rate.
        kernal: 'global': global alignment kernal, 'simple':simple assembly
    """
    base_n = len(alphabeta)
    concensus = np.zeros([base_n, 1000])
    pos = 0
    length = 0
    census_len = 1000
    for indx, bpread in enumerate(bpreads):
        if indx == 0:
            add_count(concensus, 0, bpread)
            continue
        prev_bpread = bpreads[indx - 1]
        if kernal == 'simple':
            disp,log_p = simple_assembly_kernal(bpread,prev_bpread,error_rate,jump_step_ratio)
        elif kernal == 'global':
            disp = global_alignment_kernal(bpread,prev_bpread,base_n)
        elif kernal == 'glue':
            disp = glue_kernal(bpread,prev_bpread)
        elif kernal == 'stick':
            disp = stick_kernal(bpread,prev_bpread)
        if disp + pos + len(bpreads[indx]) > census_len:
            concensus = np.lib.pad(concensus, ((0, 0), (0, 1000)),
                                   mode='constant', constant_values=0)
            census_len += 1000
        add_count(concensus, pos + disp, bpreads[indx])
        pos += disp
        length = max(length, pos + len(bpreads[indx]))
    return concensus[:, :length]

def match_blocks(alignment):
    tmp_start = -1 
    blocks = []
    pos_0 = 0
    pos_1 = 0
    for idx,base in enumerate(alignment[0]):
        if (alignment[0][idx] == '-') or (alignment[1][idx] == '-'):
            if tmp_start >= 0:
                blocks.append([idx - tmp_start,pos_0,pos_1])
                tmp_start = -1
        else:
            if tmp_start == -1:
                tmp_start = idx
        if alignment[0][idx] != '-':
            pos_0 += 1
        if alignment[1][idx] != '-':
            pos_1 += 1
    if tmp_start >=0:
        blocks.append([idx - tmp_start,pos_0,pos_1])
    return blocks

def add_count(concensus, start_indx, segment):
    if start_indx < 0:
        segment = segment[-start_indx:]
        start_indx = 0
    for i, base in enumerate(segment):
        concensus[base][start_indx + i] += 1


###############################################################################

#########################Simple assembly method with quality score################################
def simple_assembly_qs(bpreads, qs_list, jump_step_ratio,error_rate = 0.2,kernal = 'global',alphabeta = 'ACGT'):
    """
    Assemble the read from the chunks. Log probability is 
    log_P ~ x*log((N*n1/L)) - log(x!) + Ns * log(P1/0.25) + Nd * log(P2/0.25)
    Args:
        bpreads: Input chunks.
        qs_list: Quality score logits list.
        jump_step_ratio: Jump step divided by segment length.
        error_rate: An estimating basecalling error rate.
        kernal: 'global': global alignment kernal, 'simple':simple assembly, 'glue':glue assembly, 'stick':stick assembly
    """
    base_n = len(alphabeta)
    concensus = np.zeros([base_n, 1000])
    concensus_qs = np.zeros([base_n, 1000])
    pos = 0
    length = 0
    census_len = 1000
    assert len(bpreads) == len(qs_list)
    for indx, bpread in enumerate(bpreads):
        if indx == 0:
            add_count_qs(concensus, concensus_qs, 0, bpread, qs_list[indx])
            continue
        prev_bpread = bpreads[indx - 1]
        if kernal == 'simple':
            disp,log_p = simple_assembly_kernal(bpread,prev_bpread,error_rate,jump_step_ratio)
        elif kernal == 'global':
            disp = global_alignment_kernal(bpread,prev_bpread)
        elif kernal == 'glue':
            disp = glue_kernal(bpread,prev_bpread)
        elif kernal == 'stick':
            disp = stick_kernal(bpread,prev_bpread)
        if disp + pos + len(bpread) > census_len:
            concensus = np.lib.pad(concensus, ((0, 0), (0, 1000)),
                                   mode='constant', constant_values=0)
            concensus_qs = np.lib.pad(concensus_qs, ((0, 0), (0, 1000)),
                                      mode='constant', constant_values=0)
            census_len += 1000
        add_count_qs(concensus, concensus_qs, pos + disp, bpread, qs_list[indx])
        pos += disp
        length = max(length, pos + len(bpread))
    return concensus[:, :length], concensus_qs[:, :length]


def add_count_qs(concensus, concensus_qs, start_indx, segment, qs):
    if start_indx < 0:
        segment = segment[-start_indx:]
        start_indx = 0
    for i, base in enumerate(segment):
        concensus[base][start_indx + i] += 1
        concensus_qs[base][start_indx + i] += qs[i]


###############################################################################

def main():
    # bpreads = ['AAGGCCTAGCT','AGGCCTAGCAA','GGCCTAGCTC','AAAGGCCTAGT']
    #    logits_sample = np.load('/home/haotianteng/UQ/deepBNS/Chiron_Project/chiron_fastqoutput/chiron/utils/logits_sample.npy')
    start = time.time()
    # test = mc_path(logits_sample[300,:,:],base_type = 0)
    # print time.time()-start
#    bpreads = section_decoding(logits_sample)
# census = simple_assembly(bpreads)

#    result = np.argmax(census,axis=0)
#    print result



#Initialize xron package, need to run this when first time runnning xron
import os
import wget
import xron
import zipfile
MODEL_URL="https://xronmodel.s3.us-east-1.amazonaws.com/models.zip"
MODEL_PATH=xron.__path__[0]+"/models"
def get_models(args):
    if not os.path.exists(MODEL_PATH):
        os.makedirs(MODEL_PATH)
    print("Downloading models...")
    wget.download(MODEL_URL, out=MODEL_PATH)
    print("\nExtracting models...")
    with zipfile.ZipFile(MODEL_PATH+"/models.zip", 'r') as zip_ref:
        zip_ref.extractall(MODEL_PATH)
    os.remove(MODEL_PATH+"/models.zip")
    print("Done!")



    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 22:09:32 2022

@author: heavens
"""
import os
import re
import sys
import h5py
import numpy as np
import argparse
import pathlib
from tqdm import tqdm
from xron import __version__
FILE = pathlib.Path(__file__).resolve()
FLAGS = {"5mC":"C+m","5hmC":"C+h","5fC":"C+f","5caC":"C+c",
         "5hmU":"T+g","5fU":"T+e","5caU":"T+b",
         "6mA":"A+a",
         "8oxoG":"G+o",
         "Xao":"N+n"}
def check_index(fastq):
    if not os.path.exists(fastq+'.index'):
        raise FileNotFoundError('fastq index file not found')

def read_index(fastq):
    check_index(fastq)
    index_dict = {}
    with open(fastq+'.index') as f:
        index = f.read().splitlines()
        index_dict[index[0]] = index[1]
    return index_dict

def int8_encode(x):
    assert np.all(0 <= x) and np.all(x<= 1)
    return (x*256).astype(np.uint8)

class SamParser(object):
    def __init__(self,sam_file,fastq_index,modified,format='merge'):
        self.sam_file = sam_file
        self.fastq_index = fastq_index
        self.tagged = False #To identify if the file has been tagged by xron before.
        self.modified = modified
        self.flag = FLAGS[modified]
        self.canonical_base = self.flag.split('+')[0]
        self.format = format
    @property
    def PG_header(self):
        return '@PG\tID:xron_tagging\tPN:xron\tVN:%s\tCL:%s' % (__version__, 'python '+str(FILE)+' '+' '.join(sys.argv[1:]))

    def _remove_exisiting_xron_header(self):
        for header_line in self.headers:
            if 'ID:xron_tagging' in header_line:
                print("Found existing xron header, update it")
                self.headers.remove(header_line)
                self.tagged = True

    def add_header(self):
        self._remove_exisiting_xron_header()
        for i,h in enumerate(self.headers):
            #Edit the PP tag in the first PG header to create chain
            if h.startswith('@PG') and "PP:xron_tagging" not in h:
                if "PP:" in h:
                    raise ValueError("PP tag already exists in the first PG header")
                split_h = h.split('\t')
                h =  split_h[:2] + ['PP:xron_tagging'] + split_h[2:]
                h = '\t'.join(h)
                self.headers[i] = h
                break
        self.headers = [self.PG_header] + self.headers
    
    def read_fastq_index(self):
        read_ids, read_mappings = [],[]
        with open(self.fastq_index,'r') as f:
            for line in f:
                split_line = line.strip().split('\t')
                read_ids.append(split_line[0])
                read_mappings.append(split_line[1])
        self.read_ids = np.asarray(read_ids)
        self.read_mappings = np.asarray(read_mappings)

    def read_sam(self):
        with open(self.sam_file,'r') as f:
            sam = f.read().splitlines()
        self.headers,self.alignment = self._parse_header(sam)
        self.alignment_ids = [x.split('\t')[0] for x in self.alignment]
    
    def _parse_header(self,sam):
        return [line for line in sam if line.startswith('@')],[line for line in sam if not line.startswith('@')]

    def parse_MMflag(self,aln):
        split_aln = aln.strip().split('\t')
        MMs = {}
        MLs = {}
        for field in split_aln:
            if field.startswith('MM:Z:'):
                MM_tags = field[5:].strip(';').split(';')
                for MM_tag in MM_tags:
                    codes,positions = self.parse_MMtag(MM_tag)
                    for code,position in zip(codes,positions):
                        MMs[code] = position
            if field.startswith("ML:B:C"):
                split_ML = field[7:].split(',')
                assert len(split_ML) == sum(len(x.split(',')) for x in MMs.values())
                for key,val in MMs.items():
                    c = len(val.split(','))
                    MLs[key] = split_ML[:c]
                    del split_ML[:c]
        return MMs,MLs

    def parse_MMtag(self,MM_tag):
        #Parse single MM tag for each modification
        MM_tag_split = MM_tag.split(',')
        title = MM_tag_split[0].strip('.')
        unmo_base,mo_base = title.split('+')
        mo_codes,mo_pos = [],[]
        for m in mo_base:
            mo_codes.append(unmo_base+'+'+m)
            mo_pos.append(','.join(MM_tag_split[1:]))
        return mo_codes,mo_pos

    
    def _remove_exisiting_MMrecord(self,MMs,MLs):
        if self.flag in MMs.keys():
            del MMs[self.flag]
        if self.flag in MLs.keys():
            del MLs[self.flag]
    
    def _generate_MMtag(self,int8_modified_probability):
        #Generate MM tag for each modification
        if len(int8_modified_probability) == 0:
            return '',''
        pos = np.where(int8_modified_probability)[0]
        shift = pos[1:] - pos[:-1]
        shift -= 1
        if len(pos):
            gap_count = np.append([pos[0]],shift)
        else:
            gap_count = []
        mm = ','.join([str(x) for x in gap_count])
        ml = ','.join([str(x) for x in int8_modified_probability[pos]])
        return mm,ml
    
    def _generate_MMtag_nomerge(self,int8_modified_probability):
        #Generate MM tag with out merging 0 probability.
        if len(int8_modified_probability) == 0:
            return '',''
        mm = ','.join(['0']*len(int8_modified_probability))
        ml = ','.join([str(x) for x in int8_modified_probability])
        return mm,ml

    def read_modified(self,fast5_read_handle):
        modified_probability = None
        for entry in fast5_read_handle['Analyses'].keys():
            if not entry.startswith('Basecall_'):
                continue #Skip non-basecall entries
            if fast5_read_handle['Analyses'][entry].attrs['name'] != 'Xron':
                continue
            result_h = fast5_read_handle['Analyses'][entry]['BaseCalled_template']
            try:
                seq = str(np.asarray(result_h['Fastq']).astype(str)).split('\n')[1]
            except:
                seq = np.asarray(result_h['Fastq']).tobytes().decode('utf-8').split('\n')[1]
            try:
                modified_probability = np.asarray(result_h['ModifiedProbability'])
                if (len(modified_probability) == 0) and (seq.count(self.canonical_base)):
                    return None
                return int8_encode(modified_probability)
            except KeyError:
                continue
        return None

    def __call__(self):
        self.read_sam()
        self.read_fastq_index()
        self.add_header()
        fail_count = 0
        flag = self.flag
        mmtag_func = self._generate_MMtag if self.format == 'merge' else self._generate_MMtag_nomerge
        new_sam = self.headers
        uniq_file_list = set(self.read_mappings)
        with tqdm() as t:
            for fast5f in uniq_file_list:
             with h5py.File(fast5f,'r') as root:
                for read_id in self.read_ids[self.read_mappings==fast5f]:
                    try:
                        idx = self.alignment_ids.index(read_id)
                    except ValueError:
                        fail_count +=1
                        continue
                    aln = self.alignment[idx]
                    MMs,MLs = self.parse_MMflag(aln)
                    self._remove_exisiting_MMrecord(MMs, MLs)
                    read_id = aln.split('\t')[0]
                    t.postfix = "Read_id: %s, failed: %d"%(read_id,fail_count)
                    t.update()
                    read_h = root['read_' + read_id]
                    modified_p = self.read_modified(read_h)
                    if modified_p is None:
                        fail_count += 1
                    else:
                        mm,ml = mmtag_func(modified_p)
                        MMs[self.flag] = mm
                        MLs[self.flag] = ml
                    if len(MMs):
                        MM_string = 'MM:Z:'+';'.join([k+'.,'+v for k,v in MMs.items()])+';'
                        ML_string = 'ML:B:C,'+','.join([v for v in MLs.values()])
                        if 'MM:Z:' in aln:
                            aln = re.sub('MM:Z:.*?\t',MM_string+'\t',aln)
                        else:
                            aln = aln + '\t' + MM_string +'\t'
                        if 'ML:B:C' in aln:
                            aln = re.sub('ML:B:C.*?;',ML_string+';',aln)
                        else:
                            aln = aln + ML_string+';'
                        new_sam.append(aln)
        new_sam_file = os.path.splitext(self.sam_file)[0] + ".tagged.sam"
        with open(new_sam_file + '','w') as f:
            f.write('\n'.join(new_sam))

def main(args):
    if not args.fastq.endswith('.index'):
        args.fastq += '.index'
    if args.merge:
        format = "merge"
    else:
        format = "flatten"
    sam_writer = SamParser(args.sam,args.fastq,args.modified,format = format)
    sam_writer.read_sam()
    sam_writer()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Add modification tag into sam file.')
    parser.add_argument('--fastq', required = True, type=str, help='The merged fastq file')
    parser.add_argument('--sam', required = True, type=str, help='The sam file')
    parser.add_argument('--modified',default = "6mA", type=str, help='The modified base, \
                                                can be one of the %s'%list(FLAGS.keys()))
    parser.add_argument('--merge',action = "store_true",dest = "merge",
                        help = "Set the output MM tag format to compact format.")
    args = parser.parse_args()
    main(args)



"""
Created on Tue Jun 28 02:32:18 2022

@author: Haotian Teng
"""
import toml
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from xron.nrhmm.hmm_relabel import get_effective_kmers


chunk_control = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_xron_4000_noise/chunks.npy")
chunk_m6A = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/m6A/rep1/kmers_guppy_4000_dwell/chunks.npy")
kmer_m6A = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/m6A/rep1/kmers_guppy_4000_dwell/kmers.npy")
chunk_control = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_dwell/chunks.npy")
chunk_control_renorm = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_dwell/chunks_renorm.npy")
kmer_control = np.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/control/rep1/kmers_guppy_4000_dwell/kmers.npy")
chunk_control_NA = np.load("/home/heavens/bridge_scratch/NA12878_RNA_IVT/guppy_train/kmers_guppy_4000_dwell/chunks.npy")
config = toml.load("/home/heavens/bridge_scratch/ime4_Yearst/IVT/m6A/rep1/kmers_guppy_4000_dwell/config.toml")
chunk_control_xron = np.load("/home/heavens/bridge_scratch/NA12878_RNA_IVT/xron_output/kmers_guppy_4000_dwell/chunks_renorm.npy")
chunk_m6A90 = np.load("/home/heavens/bridge_scratch/m6A_Nanopore_RNA002/data/m6A_90_pct/20210430_1745_X2_FAQ15454_23428362/kmers_guppy_4000_dwell/chunks.npy")
idx2kmer = config['idx2kmer']
effective_kmers = get_effective_kmers("!MA",idx2kmer)

figs,axs = plt.subplots(ncols = 3,sharey = True,figsize = (10,5))
sns.distplot(np.mean(chunk_m6A_renorm,axis = 1),ax = axs[0],label = "m6A-Eva")
sns.distplot(np.mean(chunk_control_renorm,axis = 1),ax = axs[0],label = "control-Eva")
axs[0].legend()
sns.distplot(np.mean(chunk_m6A_renorm,axis = 1),ax = axs[1],label = "m6A-Eva")
sns.distplot(np.mean(chunk_control_xron,axis = 1),ax = axs[1],label = "control-NA12878")
axs[1].legend()
sns.distplot(np.mean(chunk_m6A90,axis = 1),ax = axs[2],label = "m6A-NA12878")
sns.distplot(np.mean(chunk_control_xron,axis = 1),ax = axs[2],label = "control-xron-NA12878")
axs[2].legend()
plt.legend()


"""
@author: heavens
"""
from torch import nn
from torch import Tensor
import numpy as np
import torch
import xron.nn
import pandas as pd
from copy import deepcopy
from functools import partial
from itertools import permutations
from fast_ctc_decode import beam_search, viterbi_search
from numpy.random import default_rng
from xron.utils.vq import vq
from xron.utils.conformer import ConformerLayer
import editdistance
from typing import List,Dict,Tuple

### Encoder Configuration
module_dict = {'Res1d':partial(xron.nn.Res1d,batch_norm = nn.LayerNorm),
               'Conformer':ConformerLayer,
               'BidirectionalRNN':xron.nn.BidirectionalRNN,
               'Attention_norm':xron.nn.AttentionNormalize,
               'LSTM':nn.LSTM,
               'ReLU':nn.ReLU,
               'GELU':nn.GELU,
               'GRU':nn.GRU,
               'Sigmoid':nn.Sigmoid,
               'BatchNorm':nn.BatchNorm1d,
               'LayerNorm':nn.LayerNorm}
# PORE_MODEL_F = "pore_models/5mer_level_table.model"
PORE_MODEL_F = "pore_models/m6A_5mer_level.model"
N_BASE = 5 #AGCTM
EMBEDDING_SIZE = 128
CNN_KERNAL_COMP= 25
CNN_STRIDE = 11
class CNN_CONFIG(object):   
    CNN = {'N_Layer':3,
           'Layers': [{'layer_type':'Res1d','kernel_size':5,'stride':1,'out_channels':16},
                      {'layer_type':'Res1d','kernel_size':5,'stride':1,'out_channels':32},
                      # {'layer_type':'Conformer','num_attention_heads':4,'ffn_dim':64,'depthwise_conv_kernel_size':7},
                      # {'layer_type':'Conformer','num_attention_heads':4,'ffn_dim':64,'depthwise_conv_kernel_size':7},
                      # {'layer_type':'Conformer','num_attention_heads':4,'ffn_dim':64,'depthwise_conv_kernel_size':7},
                      {'layer_type':'Res1d','kernel_size':CNN_KERNAL_COMP,'stride':CNN_STRIDE,'out_channels':768}]
        }
class RNN_CONFIG(CNN_CONFIG):
    RNN = {'layer_type':'BidirectionalRNN','hidden_size':768,'cell_type':'LSTM','num_layers':3}

class FNN_CONFIG(RNN_CONFIG):
    FNN = {'N_Layer':2,
           'Layers':[{'out_features':EMBEDDING_SIZE,'bias':True,'activation':'Sigmoid','dropout':None,'normalization':None},
                     {'out_features':N_BASE+1,'bias':True,'activation':'Linear','dropout':None,'normalization':None}]}
class CONFIG(FNN_CONFIG):
    def __init__(self):
        self.EMBEDDING= {'n_layers':self.CNN['N_Layer']+3}
        if self.EMBEDDING['n_layers'] ==  (self.CNN['N_Layer']+2):
            self.EMBEDDING['shape'] = EMBEDDING_SIZE*2
        else:
            self.EMBEDDING['shape'] = EMBEDDING_SIZE
        self.EMBEDDING['mask_cnn_ratio'] = 0.
            
### Critic Configuration
class CRITIC_CNN(object):
    CNN = {'N_Layer':3,
           'Layers': [{'layer_type':'Res1d','kernel_size':5,'stride':1,'out_channels':32},
                      {'layer_type':'Res1d','kernel_size':5,'stride':1,'out_channels':32},
                      {'layer_type':'Res1d','kernel_size':5,'stride':1,'out_channels':32}]
        }

class CRITIC_RNN(CRITIC_CNN):
    RNN = {'layer_type':'BidirectionalRNN','hidden_size':64,'cell_type':'LSTM','num_layers':1}
class CRITIC_FNN(CRITIC_RNN):
    FNN = {'N_Layer':2,
           'Layers':[{'out_features':32,'bias':True,'activation':'ReLU','dropout':0.2},
                     {'out_features':1,'bias':False,'activation':'Linear','dropout':None}]}
class CRITIC_CONFIG(CRITIC_FNN):
    pass


### Decoder configuration
module_dict['RevRes1d'] = xron.nn.RevRes1d

class DECODER_CONFIG(CONFIG):
    def __init__(self):
        super().__init__()
        self.CNN_DECODER = {'N_Layer':3,
                            'Layers': [{'layer_type':'RevRes1d','kernel_size':CNN_KERNAL_COMP,'stride':CNN_STRIDE,'out_channels':32},
                                       {'layer_type':'RevRes1d','kernel_size':5,'stride':1,'out_channels':32},
                                       {'layer_type':'RevRes1d','kernel_size':5,'stride':1,'out_channels':32}]
                            }
        self.CNN_DECODER['Input_Shape'] = self.EMBEDDING["shape"]
    FNN_DECODER = {'N_Layer':2,
                   'Layers':[{'out_features':8,'bias':True,'activation':'ReLU'},
                             {'out_features':1,'bias':True,'activation':'Linear'}]}
        
class MM_CONFIG(DECODER_CONFIG):
    PORE_MODEL = {"PORE_MODEL_F":PORE_MODEL_F,
                  "N_BASE": N_BASE, #for AGCT is 4, for AGCTM is 5
                  'K' : 3,
                  'N_EMBD':256, #Number of embbding, if None then it's set to N_BASE**K+1
                  "EMBEDDING_SIZE":EMBEDDING_SIZE,
                  "LOAD": False} #If load the pretrain pore model.
    DECODER = {"X_UPSAMPLING":5, #The scale factor of upsampling.
               "USE_STD":False}

def copy_config(config):
    class CONFIG:
        pass
    config_copy = CONFIG()
    config_modules = [x for x in config.__dir__() if not x .startswith('_')][::-1]
    for module in config_modules:
       setattr(config_copy,module,deepcopy(getattr(config,module)))
    return config_copy

class BASE(nn.Module):
    """The base class define the operation constructing a NN from configuration.
    """
    
    def _make_cnn(self,cnn_config:dict,in_channels = 1):
        cnn_config = deepcopy(cnn_config)
        layers = []
        for l in cnn_config['Layers']:
            layer_type = l.pop('layer_type')
            block = module_dict[layer_type]
            if layer_type == "Conformer":
                layers.append(xron.nn.Permute([0,2,1]))
            layers.append(block(in_channels = in_channels,**l))
            if layer_type == "Conformer":
                layers.append(xron.nn.Permute([0,2,1]))
            else:
                #For Conformer out_channels = in_channels
                in_channels = l['out_channels']
        return layers
    
    def _make_rnn(self,rnn_config,in_channels):
        rnn_config = deepcopy(rnn_config)
        block = module_dict[rnn_config.pop('layer_type')]
        cell = module_dict[rnn_config.pop('cell_type')]
        return [block(input_size = in_channels,
                            cell = cell,
                            **rnn_config)]
    
    def _make_fnn(self,fnn_config,in_channels):
        fnn_config = deepcopy(fnn_config)
        layers = []
        for l in fnn_config['Layers']:
            activation = l.pop('activation')
            if "dropout" in l:
                dropout = l.pop('dropout')
            else:
                dropout = None
            if "normalization" in l:
                normalization = l.pop('normalization')
            else:
                normalization = None
            layers.append(nn.Linear(in_features = in_channels,**l))
            if normalization is not None:
                layers.append(module_dict[normalization](l['out_features']))
            if activation != 'Linear':
                layers.append(module_dict[activation]())
            if dropout is not None:
                layers.append(nn.Dropout(p = dropout))
            in_channels = l['out_features']
        return layers
    
    def forward(self,batch):
        return self.net(batch)

class CRITIC(BASE):
    def __init__(self,config:CONFIG):
        """
        A Convolutional-Recurrent neural network for encoding the signal into
        base probability.

        Parameters
        ----------
        config : CONFIG
            The configuration used to generate the net, need to contain CNN,
            RNN and FNN attributes.
        """
        super().__init__()
        self.config = config
        cnn = self._make_cnn(config.CNN)
        permute = xron.nn.Permute([2,0,1]) #[N,C,L] -> [L,N,C]
        rnn = self._make_rnn(config.RNN,
                             in_channels = config.CNN['Layers'][-1]['out_channels'])
        directions = 2 if self.config.RNN['layer_type'] == "BidirectionalRNN" else 1
        fnn = self._make_fnn(config.FNN.copy(),
                             in_channels = config.RNN['hidden_size']*directions)
        self.net = nn.Sequential(*cnn,permute,*rnn,*fnn)
        self.mse_loss = nn.MSELoss(reduction = "none")
        


class CRNN(BASE):
    def __init__(self,config:CONFIG):
        """CRNN
        A Convolutional-Recurrent neural network for encoding the signal into
        base probability.

        Parameters
        ----------
        config : CONFIG
            The configuration used to generate the net, need to contain CNN,
            RNN and FNN attributes.
        """
        super().__init__()
        self.config = config
        # att = self._make_attention_norm(config.ATTENTION, 1)
        # cnn = self._make_cnn(config.CNN,in_channels = config.ATTENTION['Layers'][-1]['hidden_num'])
        cnn = self._make_cnn(config.CNN)
        permute = xron.nn.Permute([2,0,1]) #[N,C,L] -> [L,N,C]
        rnn = self._make_rnn(config.RNN,
                             in_channels = config.CNN['Layers'][-1]['out_channels'])
        directions = 2 if self.config.RNN['layer_type'] == "BidirectionalRNN" else 1
        fnn = self._make_fnn(config.FNN.copy(),
                             in_channels = config.RNN['hidden_size']*directions)
        log_softmax = nn.LogSoftmax(dim = 2)
        # self.net = nn.Sequential(*att,*cnn,permute,*rnn,*fnn,log_softmax)
        self.net = nn.Sequential(*cnn,permute,*rnn,*fnn,log_softmax)
        self.embedding_layers = [layer for layer in self.net[:self.config.EMBEDDING['n_layers']]]
        self.ctc = nn.CTCLoss(zero_infinity = False)
        self.nn_embd = vq
        if self.config.EMBEDDING["mask_cnn_ratio"] > 0:
            self.rng = default_rng()
    
    def forward_embedding(self,batch,embedding = None, training = True):
        for layer_i,layer in enumerate(self.embedding_layers):
            batch = layer(batch)
            if self.config.EMBEDDING["mask_cnn_ratio"] > 0 and layer_i == (self.config.CNN['N_Layer']-1) and training:
                T = batch.shape[-1]
                mask_T = int(T*self.config.EMBEDDING["mask_cnn_ratio"])
                mask_idxs = self.rng.choice(T, size = mask_T, replace = False)
                batch[:,:,mask_idxs] = 0 #Mask out channels
        if embedding:
            q = batch.permute([1,2,0]) # [L,N,C] -> [N,C,L]
            sg_q = q.detach()
            e,e_shadow = self.nn_embd(batch.permute([1,0,2]),embedding.weight) #[*,*,*] -> [N,L,C]
            e = e.permute([0,2,1]) #[N,L,C] -> [N,C,L]
            e_shadow = e_shadow.permute([0,2,1]) #[N,L,C] -> [N,C,L]
            sg_e = e.detach()
            return e,q,e_shadow,sg_q,sg_e
        else:
            return batch
    
    def ctc_loss(self,
                 posterior:Tensor,
                 posterior_len:Tensor,
                 seq:Tensor,
                 seq_len:Tensor):
        return self.ctc(posterior,seq,posterior_len,seq_len)
    
    def ctc_loss_uncertain_label(self,
                                 posterior:Tensor,
                                 posterior_len:Tensor,
                                 seq:Tensor,
                                 seq_len:Tensor,
                                 uncertain_label:Tuple[int,int],
                                 uncertain_label_weight:float = 0.1):
        """
        Calculate the smooth CTC loss with label i replaced by j with weight e,
        where P(y) = e * P(y=i) + (1-e) * P(y=j)
        And the label smooth loss is calculated as:
            L = e*CTCloss(S) + (1-e)*CTCloss(S[i->j])
        Where S[i->j] is the sequence with all i replaced by j.
        """
        i,j = uncertain_label
        seq_i2j = seq.clone()
        seq_i2j[seq_i2j == i] = j
        loss = (1-uncertain_label_weight) * self.ctc(posterior,seq,posterior_len,seq_len) + \
                uncertain_label_weight * self.ctc(posterior,seq_i2j,posterior_len,seq_len)
        return loss

    def ctc_decode(self,
                   posteriors:Tensor,
                   alphabet:str = "NACGT",
                   beam_size:int = 5,
                   beam_cut_threshold:float = 0.05):
        """
        Use beam-search CTC decoder to decode the given posterior probability.

        Parameters
        ----------
        posteriors : Tensor
            A L-N-C posterior probability tensor. Where L is the signal length,
            N is the batch size and C is the number of alphabet.
        alphabet : str, optional
            A string contains the alphabet, first character is the blank symbol.
            The default is "NACGT".
        beam_size : int, optional
            The beam size, if it's 1, a faster but less-accurate vaterbi
            decoder will be used. The default is 5.
        beam_cut_threshold: float, optional
            The cut threshold of beam search, the higher it is, more search
            will done, too large the value could run out of search space of 
            beam search. The default is 0.05

        Returns
        -------
        A N-length list of the predicting sequence.

        """
        if beam_size <=1:
            decoder = partial(viterbi_search, alphabet = alphabet)
        else:
            decoder = partial(beam_search, beam_size = beam_size, alphabet = alphabet)
        posteriors = posteriors.cpu().detach().numpy()
        batch_size = posteriors.shape[1]
        predictions = []
        for i in np.arange(batch_size):
            pred = decoder(np.exp(posteriors[:,i,:]))[0]
            predictions.append(pred)
        return predictions
        
    
    def error(self,
              predictions:List[np.ndarray],
              seqs:Tensor,
              seqs_len:Tensor,
              alphabet:str = "ACGT"):
        """
        Calculate the ctc decoding error between the true label.

        Parameters
        ----------
        predictions: List[np.ndarray]
            A list of vectors, each vector is the predicting sequence.
        seqs : Tensor
            A N-L matrix contains the true sequence label.
        seqs_len : Tensor
            A N-1 vector contains the length of the sequences.
        alphabet : str, optional
            A string contains the alphabet without the blank symbol.
            Can be a permutation of the alphabet used in ctc_decode.
            The default is "ACGT".

        Returns
        -------
        None.

        """
        batch_size = len(predictions)
        seqs = seqs.cpu().detach().numpy()
        seqs_len = seqs_len.cpu().detach().numpy().flatten()
        errors = []
        for i in np.arange(batch_size):
            pred = predictions[i]
            seq_len = seqs_len[i]
            error = editdistance.eval(pred,"".join([alphabet[x-1] for x in seqs[i][:seq_len]]))
            errors.append(error/seq_len)
        return np.array(errors).mean()
    
    def ctc_error(self,
                  posteriors:Tensor,
                  seqs:Tensor,
                  seqs_len:Tensor,
                  alphabet:str = "NACGT",
                  beam_size:int = 5,
                  beam_cut_threshold:float = 0.05,
                  reduction:Dict = None):
        """
        Calcualte the CTC decoding error of the given posterior probability
        and sequence with true label.

        Parameters
        ----------
        posteriors : Tensor
            A L-N-C posterior probability tensor. Where L is the signal length,
            N is the batch size and C is the number of alphabet.
        seqs : Tensor
            A N-L matrix contains the true sequence label.
        seqs_len : Tensor
            A N-1 vector contains the length of the sequences.
        alphabet : str, optional
            A string contains the alphabet, first character is the blank symbol.
            The default is "NACGT".
        beam_size : int, optional
            The beam size, if it's 1, a faster but less-accurate vaterbi
            decoder will be used. The default is 5.
        beam_cut_threshold: float, optional
            The cut threshold of beam search, the higher it is, more search
            will done, too large the value could run out of search space of 
            beam search. The default is 0.05
        reduction: dictionary, optional
            Default is None, conduct no reduction, if a dict is provided, the
            modifed bases will be reducted to original bases, e.g. {M:A} will
            reduct M base back to A and then compare the error.

        Returns
        -------
        The average error.

        """
        if torch.sum(torch.isnan(posteriors)):
            print("Posterior has NaN in it.")
            return np.nan
        preds = self.ctc_decode(posteriors = posteriors,
                                alphabet=alphabet,
                                beam_size=beam_size,
                                beam_cut_threshold=beam_cut_threshold)
        if reduction:
            error = self.reduction_error(preds, 
                                         seqs, 
                                         seqs_len,
                                         alphabet = alphabet[1:],
                                         modified_bases = reduction)
        else:
            error = self.error(preds,seqs,seqs_len,alphabet = alphabet[1:])
        return error
    
    def reduction_error(self,
              predictions:List[np.ndarray],
              seqs:Tensor,
              seqs_len:Tensor,
              modified_bases:Dict = {'M':'A'},
              alphabet:str = "ACGTM"):
        """
        Calculate the ctc decoding error between the true label.

        Parameters
        ----------
        predictions: List[np.ndarray]
            A list of vectors, each vector is the predicting sequence.
        seqs : Tensor
            A N-L matrix contains the true sequence label.
        seqs_len : Tensor
            A N-1 vector contains the length of the sequences.
        alphabet : str, optional
            A string contains the alphabet without the blank symbol.
            Can be a permutation of the alphabet used in ctc_decode.
            The default is "ACGT".

        Returns
        -------
        None.

        """
        batch_size = len(predictions)
        seqs = seqs.cpu().detach().numpy()
        seqs_len = seqs_len.cpu().detach().numpy().flatten()
        errors = []
        for i in np.arange(batch_size):
            pred = predictions[i]
            seq_len = seqs_len[i]
            plain_seq = "".join([alphabet[x-1] for x in seqs[i][:seq_len]])
            for key,val in modified_bases.items():
                plain_seq.replace(key,val)
                palin_pred = pred.replace(key,val)
            error = editdistance.eval(palin_pred,plain_seq)
            errors.append(error/seq_len)
        return np.array(errors).mean()
    
    def permute_error(self,
                      posteriors:Tensor,
                      seqs:Tensor,
                      seqs_len:Tensor,
                      alphabet:str = "NACGT",
                      beam_size:int = 5,
                      beam_cut_threshold:float = 0.05):
        """
        Calcualte the permutating error of the given posterior probability
        and sequence with true label. Permutation of the alphabet will be used
        to calculate the minimum error given the CTC decoding result.

        Parameters
        ----------
        posteriors : Tensor
            A L-N-C posterior probability tensor. Where L is the signal length,
            N is the batch size and C is the number of alphabet.
        seqs : Tensor
            A N-L matrix contains the true sequence label.
        seqs_len : Tensor
            A N-1 vector contains the length of the sequences.
        alphabet : str, optional
            A string contains the alphabet, first character is the blank symbol.
            The default is "NACGT". All permutations of the alphabet will be tested.
        beam_size : int, optional
            The beam size, if it's 1, a faster but less-accurate vaterbi
            decoder will be used. The default is 5.
        beam_cut_threshold: float, optional
            The cut threshold of beam search, the higher it is, more search
            will done, too large the value could run out of search space of 
            beam search. The default is 0.05

        Returns
        -------
        A list of errors with the permutation.

        """
        preds = self.ctc_decode(posteriors = posteriors,
                                alphabet=alphabet,
                                beam_size=beam_size,
                                beam_cut_threshold=beam_cut_threshold)
        ab = alphabet[1:]
        errors = []
        perms = []
        for perm in permutations(ab):
            perm = ''.join(perm)
            error = self.error(preds,seqs,seqs_len,alphabet = perm)
            errors.append(error)
            perms.append(perm)
        return np.asarray(errors), np.asarray(perms)

class REVCNN(BASE):
    def __init__(self,config:DECODER_CONFIG):
        """
        The reverse convolutional neural network to reconstruct the signal from
        the base probability, the decoder.

        Parameters
        ----------
        config : DECODER_CONFIG
            A Configuration used to generate this net, need to contain 
            CNN_DECODER and FNN_DECODER attributes. Input should be a tensor 
            with shape [N,C,L], output is a tensor with shape [N,L,C]

        Returns
        -------
        None.

        """
        super().__init__()
        self.config = config
        self.embedding_size = config.CNN_DECODER['Input_Shape']
        cnn = self._make_cnn(config.CNN_DECODER, in_channels = self.embedding_size)
        permute = xron.nn.Permute([0,2,1]) #[N,C,L] -> [N,L,C]
        fnn = self._make_fnn(config.FNN_DECODER.copy(),
                             in_channels = config.CNN_DECODER['Layers'][-1]['out_channels'])
        self.net = nn.Sequential(*cnn,permute,*fnn)
        self.mse_loss = nn.MSELoss(reduction = 'none')
        self.entropy_loss = nn.CrossEntropyLoss()

class MM(nn.Module):
    def __init__(self,config:MM_CONFIG):
        """
        The statistical Markov generative pore model, 

        Parameters
        ----------
        config : DECODER_CONFIG
            A config contains the pore model.

        Returns
        -------
        None.
        """
        super().__init__()
        self.config = config
        self.N_BASE = self.config.PORE_MODEL['N_BASE']
        self.K = self.config.PORE_MODEL['K']
        self.n_embd = self.config.PORE_MODEL['N_EMBD']
        self.embedding_size = self.config.PORE_MODEL["EMBEDDING_SIZE"]
        if config.PORE_MODEL['LOAD']:
            self.pore_model = pd.read_csv(config.PORE_MODEL['PORE_MODEL_F'],
                                          delimiter = '\t')
            self.padding_signal = np.mean(self.pore_model.level_mean)
            level_tensor = torch.tensor([self.padding_signal]+list(self.pore_model.level_mean)).unsqueeze(1).to(torch.float32)
            self.level_embedding = nn.Embedding.from_pretrained(level_tensor,
                                                                freeze = False,
                                                                padding_idx=0)
        else:
            if not self.n_embd:
                n_embd = (self.N_BASE)**self.K+1
            else:
                n_embd = self.n_embd
            self.level_embedding = nn.Embedding(n_embd,self.embedding_size)
            # if self.config.EMBEDDING['n_layers'] ==  (self.config.CNN['N_Layer']+4):
            #     shift = 0.5 #The embedding output is from a sigmoid layer, so it's approximately sigmoid(0) ~ 0.5
            # else:
            #     shift = 0.
            self.level_embedding.weight.data.uniform_(-1./n_embd, 1./n_embd)
        self.upsampling = torch.nn.Upsample(scale_factor=config.DECODER['X_UPSAMPLING'],
                                     mode = 'nearest')
        self.entropy_loss = nn.CrossEntropyLoss()
        self.bn = torch.nn.BatchNorm1d(1)
        
    def forward(self, sampling:torch.Tensor,device = None):
        sampling = sampling.cpu().detach().numpy()
        sampling = np.argmax(sampling,axis = 1)
        kmer_batch = self._kmer_decode(sampling)
        if device:
            kmer_batch = kmer_batch.to(device = device)
        rc_signal = self.level_embedding(kmer_batch).squeeze(2).unsqueeze(1) # (N,L,1) -> (N,1,L)
        rc_signal = self.bn(self.upsampling(rc_signal))
        return rc_signal.permute([0,2,1])
        
    def _kmer_decode(self,
                     sequence_batch):
        kmer_batch = np.zeros((sequence_batch.shape[0],sequence_batch.shape[1]))
        K = self.K
        for i,sequence in enumerate(sequence_batch):
            kmer_seq = []
            curr_kmer = ''
            pre_base = 0
            for base in sequence:
                if base == 0 or base == pre_base:
                    curr_kmer = curr_kmer
                else:
                    curr_kmer = (curr_kmer + str(base))[-K:]
                pre_base = base
                kmer_seq.append(self.kmer2idx(curr_kmer))
            kmer_seq = np.asarray(kmer_seq)
            kmer_batch[i,:] = kmer_seq+1
        # signal_batch = torch.from_numpy(signal_batch[:,None,:])
        if kmer_batch.max() >= self.level_embedding.weight.shape[0]:
            print(kmer_batch.max())
            print(self.level_embedding.weight.shape)
        return torch.LongTensor(kmer_batch)
    
    def kmer2idx(self,kmer:str):
        "Map the kmer to index, e.g. AAAAA -> 0, AAAAC -> 1, AAAAG -> 2, ..."
        if len(kmer)<self.K:
            return -1
        multi = 1
        idx = 0
        for base in kmer[::-1]:
            idx += (int(base)-1)*multi
            multi = multi * self.N_BASE
        return idx
    
    def idx2kmer(self,idx:int,ab_dict = None):
        if not ab_dict:
            ab_dict = {i:str(i+1) for i in np.arange(self.N_BASE)}
        kmer = ''
        while idx>=self.N_BASE:
            kmer = ab_dict[idx%self.N_BASE] + kmer
            idx //= self.N_BASE
        kmer = ab_dict[idx%self.N_BASE] + kmer
        kmer = ab_dict[0]*(self.K-len(kmer)) + kmer              
        return kmer
            
    @property
    def mse_loss(self):
        if self.config.DECODER['USE_STD']:
            pass #TODO: Implement the std version.
        else:
            return nn.MSELoss(reduction= 'none')
    
    def save_embedding(self,alphabeta = 'ACGTM'):
        embedding = self.level_embedding.weight.detach().cpu().squeeze(1).numpy()
        ab_dict = {i:x for i,x in enumerate(alphabeta)}
        with open(self.config.PORE_MODEL['PORE_MODEL_F'],'w+') as f:
            f.write("kmer\tlevel_mean\n")
            for i,e in enumerate(embedding[1:]):
                f.write("%s\t%.6f\n"%(self.idx2kmer(i,ab_dict = ab_dict),e))
    
if __name__ == "__main__":
    config = CONFIG()
    encoder = CRNN(config)
    batch_size = 88
    chunk_len = 2000
    test_batch = torch.randn(batch_size,1,chunk_len)
    embedding = encoder.forward_embedding(test_batch)
    print("Input: ",test_batch.shape,"Encoded: ",embedding.shape)
    decoder_config = DECODER_CONFIG()
    decoder = REVCNN(decoder_config)
    rc = decoder.forward(embedding.permute([1,2,0])).permute([0,2,1])
    print("Reconstructed:",rc.shape)
    # output = encoder.forward(test_batch)
    # mm_config = MM_CONFIG()
    # mm_decoder = MM(mm_config)
    # mm_rc = mm_decoder.forward(output.permute([1,2,0]))
    # print("MM Reconstructed:",mm_rc.shape)
    # mm_decoder.save_embedding()
